from Bio.PDB.Structure import Structure as BioStructure

from luna.pdb.core.entity import Entity


class Structure(BioStructure, Entity):
    """
    Custom Structure subclass for macromolecular structures.

    Supports SMCRA traversal and stores the original file path.

    Parameters
    ----------
    id : str
        Structure identifier.
    pdb_file : str, optional
        Path to the original PDB file.
    """

    def __init__(self, id, pdb_file=None):
        super().__init__(id)
        
        self.pdb_file = pdb_file

    def __repr__(self):
        return f"<Structure (LUNA subclass) id={self.id}>"



import string
import random
import inspect
import warnings

import logging
logger = logging.getLogger()


def new_random_string(size=32, chars=string.ascii_uppercase + string.digits):
    """Generate a new random string of size ``size`` containing only the
    characters provided in ``chars``.

    Parameters
    ----------
    size : int, optional
            The size of the new string. The default value is 32.
    chars : iterable, optional
            A sequence of characters to choose from. The default value is
            'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'.

    Returns
    -------
    random_string : str
            The new random string.
    """
    return ('').join((random.choice(chars) for i in range(size)))


def rgb2hex(r, g, b):
    return "#{:02x}{:02x}{:02x}".format(r, g, b)


def hex2rgb(hexcode):
    return tuple(map(ord, hexcode[1:].decode('hex')))


# TODO: Transform hex to rgb and vice versa
class ColorPallete:
    def __init__(self, color_map=None, default_color=(255, 255, 255)):
        self.color_map = color_map or {}
        self.default_color = default_color

    def add_color(self, key, color):
        self.color_map[key] = color

    def get_normalized_color(self, key):
        color = self.get_color(key)

        if isinstance(color, str):
            return color
        elif all([x < 1 for x in color]):
            return color
        return tuple(c / 255 for c in color)

    def get_unnormalized_color(self, key):
        color = self.get_color(key)

        if isinstance(color, str):
            return color
        elif all([x > 1 for x in color]):
            return color
        return tuple(int(c * 255) for c in color)

    def get_color(self, key):
        if key in self.color_map:
            return self.color_map[key]
        elif self.default_color:
            logger.warning("Key '%s' does not exist. Then, the default color "
                           "will be used: %s." % (key, self.default_color))
            return self.default_color
        else:
            raise KeyError("Key '%s' does not exist and no default "
                           "color was defined." % key)

    def __contains__(self, key):
        return key in self.color_map


def func_call_to_str(func, *args, **kwargs):
    arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]
    args = args[:len(arg_names)]
    defaults = func.__defaults__ or ()
    args = (args + defaults[len(defaults) - (func.__code__.co_argcount - len(args)):])
    params = zip(arg_names, args)
    args = args[len(arg_names):]
    if args:
        params.append(('args', args))
    if kwargs:
        params.append(('kwargs', kwargs))

    args_as_str = ', '.join('%s=%r' % p for p in params) + ' )'
    func_as_str = "%s(%s)" % (func.__name__, args_as_str)

    return func_as_str


def iter_to_chunks(l, n):
    return [l[i:i + n] for i in range(0, len(l), n)]


class LUNAWarning(Warning):
    """Base LUNA warning class.

    Unlike normal warnings, these are by default always set to on.
    """


class LUNADeprecationWarning(LUNAWarning, DeprecationWarning):
    """A warning class for a deprecated method or class."""


# Always show custom warnings for this package
warnings.filterwarnings("always", category=LUNAWarning)


class deprecated(object):
    """Decorator to mark a function as deprecated.

    Issue a deprecation warning when a function is called, and update the
    documentation. A deprecation version must be provided.

    Examples
    --------
    >>> from luna.util import deprecated
    >>> @deprecated("1.1", remove_version="1.3",
    ...             msg="Function no longer needed")
    ... def my_function():
    ...     pass

    Notes
    -----
    Adapted from https://wiki.python.org/moin/PythonDecoratorLibrary
    """

    def __init__(self, deprecated_version, remove_version=None, msg=None):
        """Constructor.

        Parameters
        ----------
        deprecated_version : str
            Version in which object was deprecated (e.g. '1.1')
        remove_version : str, optional
            Version in which object will be removed (e.g. '1.2'). If not
            specified, it is assumed the object will be removed in the next
            release (e.g. '1.2' if `deprecated_version` is '1.1')
        msg : str, optional
            Message to include with deprecation warning, to explain deprecation
            or point to newer version.
        """
        self.deprecated_version = deprecated_version
        if remove_version is None:
            version_info = deprecated_version.split(".")
            version_info[1] = str(int(version_info[1]) + 1)
            for i in range(2, len(version_info)):
                version_info[i] = "0"
            remove_version = ".".join(version_info)
        self.remove_version = remove_version
        if msg is None:
            self.extra = ""
        else:
            self.extra = " {0}".format(msg)

    def __call__(self, obj):
        if inspect.isfunction(obj):
            return self.deprecate_function(obj)
        else:
            raise ValueError("Deprecated object is not a function.")

    def deprecate_function(self, f):
        """Return the decorated function."""
        msg = (
            "Function `{0}` was deprecated in {1} and will be removed "
            "in {2}.{3}"
        ).format(
            f.__name__,
            self.deprecated_version,
            self.remove_version,
            self.extra,
        )

        def new_func(*args, **kwargs):
            warnings.warn(msg, category=LUNADeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)

        new_func.__name__ = f.__name__
        new_func.__dict__ = f.__dict__
        new_func.__doc__ = f.__doc__
        self.update_docstring(new_func)
        return new_func

    def update_docstring(self, obj):
        """Add deprecation note to docstring."""
        msg = (
            ".. note:: Deprecated in LUNA {0}.\n"
            "   `{1}` will be removed in LUNA {2}.{3}"
        ).format(
            self.deprecated_version,
            obj.__name__,
            self.remove_version,
            self.extra,
        )
        obj.__doc__ = "{0}\n\n{1}".format(msg, obj.__doc__)
        return obj



from enum import Enum, auto


class IFPType(Enum):
    """An enumeration of Interaction FingerPrints (IFPs) available in LUNA."""

    # auto() creates automatic identifiers for each type as this value is
    # not important. Therefore, always remember to compare IFP types using
    # their names and not their values.

    # Extended Interaction FingerPrint
    EIFP = auto()

    # Hybrid version: atomic invariants for atoms and pharmacophore
    # for groups
    HIFP = auto()

    # Functional Interaction FingerPrint
    FIFP = auto()



from subprocess import Popen, PIPE
import tempfile
import os

from luna.wrappers.base import MolWrapper
from luna.wrappers.chemaxon.util import check_license
from luna.util.file import (new_unique_filename, is_file_valid,
                            get_file_format)
from luna.util.exceptions import (InvalidFileFormat, IllegalArgumentError,
                                  ProcessingFailed)

import logging
logger = logging.getLogger()


valid_formats = ["mol", "sdf", "sybyl", "mol2", "pdb", "xyz"]


class ConformerGenerator:

    def __init__(self, cleaner=3, skip_3d=False,
                 add_h=True, use_mmff94=True,
                 optimize=True, optim_criteria=3,
                 mol_obj_type="rdkit",
                 molconvert="molconvert"):

        self.cleaner = cleaner
        self.skip_3d = skip_3d
        self.add_h = add_h
        self.use_mmff94 = use_mmff94
        self.optimize = optimize
        self.optim_criteria = optim_criteria
        self.mol_obj_type = mol_obj_type
        self.molconvert = molconvert

    def _prep_opts(self):
        opt_list = []

        if self.cleaner in [0, 1, 2, 3]:
            if not self.skip_3d:
                opt_list += ["[c]{%d}{0}" % self.cleaner]
            else:
                opt_list += ["[c]{%d}{1}" % self.cleaner]
        else:
            raise IllegalArgumentError("The informed cleaner '%s' is invalid. "
                                       "Available options are 0, 1, 2, or 3."
                                       % str(self.cleaner))

        if self.add_h:
            opt_list += ["[prehydrogenize]"]

        if self.use_mmff94:
            opt_list += ["[mmff94]"]

        if self.optimize:
            if self.optim_criteria in [0, 1, 2, 3]:
                opt_list += ["[o]{1}{0}{%d}" % self.optim_criteria]
            else:
                raise IllegalArgumentError("The informed optimization "
                                           "criteria '%s' is invalid. "
                                           "Available options are "
                                           "0, 1, 2, or 3."
                                           % str(self.optim_criteria))
        else:
            opt_list += ["[o]{0}"]

        return opt_list

    def _run_molconvert(self, mol_input, is_smiles,
                        output_file, output_format):
        opt_list = self._prep_opts()

        input_list = [mol_input]
        if is_smiles:
            input_list = ["-s", mol_input]

        output_list = ["-o", output_file]
        args = ([self.molconvert, "-3:%s" % "".join(opt_list), output_format]
                + input_list + output_list)

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        try:
            stdout, stderr = p.communicate()
        except Exception:
            p.kill()
            raise

        error_lines = stderr.decode().strip().split("\n")

        if len(error_lines) == 1 and error_lines[0] == "":
            error_lines = []

        # The list of errors are ok if they are one of the following.
        info_msgs = ["chemaxon.license.RemoteLicenseClient",
                     "INFO: Failed to retrieve available "
                     "licenses from the server:",
                     "INFO: Failed to retrieve license file from server:"]
        critical_errors_found = False
        for e in error_lines:
            if any([m in e for m in info_msgs]):
                continue
            critical_errors_found = True
            break

        # Raise ProcessingFailed error if any critical errors were found
        # or if the output file is empty.
        if (not critical_errors_found
                and os.path.getsize(self.output_file) > 0):

            logger.debug("Structure for input '%s' created with success."
                         % mol_input)

            # Try to parse the created molecular file.
            self.mol = MolWrapper.from_mol_file(self.output_file,
                                                output_format,
                                                self.mol_obj_type)
        else:
            logger.error(stderr.decode())
            raise ProcessingFailed("molconvert could not generate the 3D "
                                   "structure. Check the logs for more "
                                   "information.")

    def generate(self, mol_input, is_smiles=False,
                 output_file=None, output_format=None):

        check_license()

        if is_smiles:
            if not isinstance(mol_input, str):
                raise TypeError("The parameter 'is_smiles' was set on. "
                                "However, an invalid input type was provided. "
                                "A string was expected instead.")

        elif isinstance(mol_input, str) and not is_file_valid(mol_input):
            raise IllegalArgumentError("'%s' does not exist or "
                                       "is not a file." % mol_input)

        elif not isinstance(mol_input, str):
            raise IllegalArgumentError("'mol_input' must be a "
                                       "string representing a molecular "
                                       "file or SMILES.")

        if isinstance(output_file, str):
            if output_format is None:
                logger.debug("The output file format was not defined. "
                             "It will assume the format by the "
                             "file extension.")
                output_format = get_file_format(output_file, 1)

            if output_format not in valid_formats:
                raise InvalidFileFormat("Output format '%s' is not valid."
                                        % output_format)

            logger.debug("Output format: %s" % output_format)

        elif output_file is not None:
            raise IllegalArgumentError("The 'output_file' should be "
                                       "a string or None.")

        if output_file is None:
            output_file = new_unique_filename(tempfile.gettempdir()) + ".mol"
            output_format = "mol"

        self.mol = None
        self.mol_input = mol_input
        self.output_file = output_file

        self._run_molconvert(mol_input, is_smiles,
                             output_file, output_format)



from os import path
from collections import defaultdict
from pdbecif.mmcif_io import CifFileReader

from luna.util.file import parse_json_file

import logging
logger = logging.getLogger()


class MoleculesData(dict):
    """Generic class to define precomputed molecular data such as
    default bonds, atomic invariants, and physicochemical
    properties.

    Parameters
    ----------
    data : dict, optional
        A dict containing molecules' data.
    """

    def __init__(self, data=None):
        data = data or {}

        super().__init__(data)

    @classmethod
    def from_json_file(cls, json_file):

        data = parse_json_file(json_file)

        return cls(data)


class DefaultResidueData(MoleculesData):
    """Default parameters for calculating interactions in LUNA."""

    def __init__(self, **kwargs):

        file_path = path.abspath(path.join(path.realpath(__file__),
                                           '../../data'))
        filename = "precomputed_residue_data.json"
        default_file = f"{file_path}/{filename}"

        data = MoleculesData.from_json_file(default_file)

        super().__init__(data)



import math
from collections import defaultdict
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from luna.mol.entry import MolFileEntry


def generate_residue_matrix(interactions_mngrs, by_interaction=True):
    """Generate a matrix to count interactions per residue.

    Parameters
    ----------
    interactions_mngrs : iterable of \
            :class:`~luna.interaction.calc.InteractionsManager`
        A sequence of :class:`~luna.interaction.calc.InteractionsManager`
        objects from where interactions will be recovered.
    by_interaction : bool
        If True (the default), count the number of each interaction type
        per residue. Otherwise, count the overall number of interactions
        per residue.

    Returns
    -------
     : :class:`pandas.DataFrame`
    """

    data_by_entry = defaultdict(lambda: defaultdict(int))
    residues = set()

    for inter_mngr in interactions_mngrs:
        entry = inter_mngr.entry

        for inter in inter_mngr:
            # Continue if no reference is in the interaction.
            if (not inter.src_grp.has_reference()
                    and not inter.trgt_grp.has_reference()):
                continue
            # Ignore interactions involving the same compounds.
            if inter.src_grp.compounds == inter.trgt_grp.compounds:
                continue

            if inter.src_grp.has_hetatm():
                comp1 = sorted(inter.src_grp.compounds)
                comp2 = sorted(inter.trgt_grp.compounds)
            elif inter.trgt_grp.has_hetatm():
                comp1 = sorted(inter.trgt_grp.compounds)
                comp2 = sorted(inter.src_grp.compounds)
            else:
                comp1 = sorted(inter.src_grp.compounds)
                comp2 = sorted(inter.trgt_grp.compounds)
                comp1, comp2 = sorted([comp1, comp2])

            comp1 = ";".join(["%s/%s/%d%s" % (r.parent.id,
                                              r.resname,
                                              r.id[1],
                                              r.id[2].strip()) for r in comp1])
            comp2 = ";".join(["%s/%s/%d%s" % (r.parent.id,
                                              r.resname,
                                              r.id[1],
                                              r.id[2].strip()) for r in comp2])

            entry_id = (entry.mol_id if isinstance(entry, MolFileEntry)
                        else entry.to_string())

            if by_interaction:
                key = (entry_id, inter.type)
            else:
                key = entry_id

            data_by_entry[key][comp2] += 1

            residues.add(comp2)

    heatmap_data = defaultdict(list)

    if by_interaction:
        entries = set([k[0] for k in data_by_entry.keys()])
        interactions = set([k[1] for k in data_by_entry.keys()])

        for e in entries:
            for i in interactions:
                for res in residues:
                    heatmap_data["entry"].append(e)
                    heatmap_data["interaction"].append(i)
                    heatmap_data["residues"].append(res)
                    heatmap_data["frequency"].append(data_by_entry[(e, i)][res])

    else:
        for key in data_by_entry:
            for res in data_by_entry[key]:
                heatmap_data["entry"].append(key)

                heatmap_data["residues"].append(res)
                heatmap_data["frequency"].append(data_by_entry[key][res])

    df = pd.DataFrame.from_dict(heatmap_data)

    if by_interaction:
        return pd.pivot_table(df,
                              index=['entry', 'interaction'],
                              columns='residues',
                              values='frequency',
                              fill_value=0)
    else:
        return pd.pivot_table(df,
                              index='entry',
                              columns='residues',
                              values='frequency',
                              fill_value=0)


def heatmap(data_df,
            figsize=None,
            cmap="Blues",
            heatmap_kw=None,
            gridspec_kw=None):
    """ Plot a residue matrix as a color-encoded matrix.

    Parameters
    ----------
    data_df : :class:`pandas.DataFrame`
        A residue matrix produced
        with :func:`~luna.analysis.residues.generate_residue_matrix`.
    figsize : tuple, optional
        Size (width, height) of a figure in inches.
    cmap : str, iterable of str
        The mapping from data values to color space.
        The default value is 'Blues'.
    heatmap_kw : dict, optional
        Keyword arguments for :func:`seaborn.heatmap`.
    gridspec_kw : dict, optional
        Keyword arguments for :class:`matplotlib.gridspec.GridSpec`.
        Used only if the residue matrix (``data_df``) contains interactions.

    Returns
    -------
     : :class:`matplotlib.axes.Axes` or :class:`numpy.ndarray` \
            of :class:`matplotlib.axes.Axes`

    """
    data_df = data_df.reset_index()

    heatmap_kw = heatmap_kw or {}
    gridspec_kw = gridspec_kw or {}

    interactions = None
    if "interaction" in data_df.columns:
        interactions = sorted(data_df["interaction"].unique())
        max_value = data_df[data_df.columns[2:]].max().max()
    else:
        max_value = data_df[data_df.columns[1:]].max().max()

    if not interactions:
        data_df.set_index('entry', inplace=True)

        fig = plt.figure(figsize=figsize)
        ax = sns.heatmap(data_df, cmap=cmap,
                         vmax=max_value, vmin=0, **heatmap_kw)
        ax.set_xlabel("")
        ax.set_ylabel("")
        return ax
    else:
        ncols = 3
        if "ncols" in gridspec_kw:
            ncols = gridspec_kw["ncols"]
            del gridspec_kw["ncols"]
        nrows = math.ceil(len(interactions) / ncols)

        fig, axs = plt.subplots(nrows, ncols, 
                                figsize=figsize, gridspec_kw=gridspec_kw)

        row, col = 0, 0
        for i, interaction in enumerate(interactions):
            df = data_df[data_df["interaction"] == interaction].copy()
            df.drop(columns="interaction", inplace=True)
            df.set_index('entry', inplace=True)

            g = sns.heatmap(df, cmap=cmap, vmax=max_value, vmin=0,
                            ax=axs[row][col], **heatmap_kw)

            g.set_title(interaction)
            g.set_xlabel("")
            g.set_ylabel("")

            col += 1
            if col == ncols:
                row += 1
                col = 0

        if len(interactions) < nrows * ncols:
            diff = (nrows * ncols) - len(interactions)
            for i in range(1, diff + 1):
                axs[-1][-1 * i].axis('off')

        return axs



import logging
from os.path import exists
from pathlib import Path
from shutil import move as rename_pdb_file

from Bio.PDB import PDBList

from luna.util.file import is_directory_valid
from luna.util.exceptions import IllegalArgumentError


logger = logging.getLogger(__name__)


PDB_SERVER = "https://files.wwpdb.org"


def download_pdb(pdb_id, 
                 output_path=".", 
                 overwrite=False,
                 server=PDB_SERVER):
    """Download a PDB file from RCSB.org.

    Parameters
    ----------
    pdb_id : str
        4-symbols structure Id from PDB (e.g. 3J92).
    output_path : str
        Directory where the PDB file should be saved.
    overwrite : bool
        If False, skip download if file already exists.
    """
    
    if not pdb_id or not pdb_id.strip():
        raise IllegalArgumentError(f"An empty PDB ID ('{pdb_id}') was provided.")

    pdb_id = pdb_id.strip()
    output_path = Path(output_path)
    
    if not is_directory_valid(output_path):
        raise IllegalArgumentError(f"Output path '{output_path}' is not valid or writable.")

    output_path.mkdir(parents=True, exist_ok=True)
    target_path = output_path / f"{pdb_id}.pdb"

    if target_path.exists() and not overwrite:
        logger.debug(f"PDB file '{target_path}' already exists. Skipping download.")
        return str(target_path)

    logger.debug(f"Downloading PDB '{pdb_id}' to '{target_path}'...")

    pdbl = PDBList(server=PDB_SERVER)
    pdbl.retrieve_pdb_file(pdb_id, 
                           pdir=output_path,
                           file_format="pdb", 
                           overwrite=overwrite)
    
    downloaded_file = output_path / f"pdb{pdb_id.lower()}.ent"
    if downloaded_file.exists():
        rename_pdb_file(str(downloaded_file), str(target_path))
        logger.debug(f"PDB '{pdb_id}' downloaded and renamed to '{target_path}'.")
    else:
        logger.warning(f"Expected file '{downloaded_file}' not found after download.")

    return str(target_path)



import numpy as np
from rdkit.DataStructs.cDataStructs import ExplicitBitVect, SparseBitVect
from scipy.sparse import issparse, csr_matrix
from collections import defaultdict
from rdkit import DataStructs

from luna.util.exceptions import (BitsValueError, InvalidFingerprintType,
                                  IllegalArgumentError, FingerprintCountsError)
from luna.version import __version__


import logging

logger = logging.getLogger()

DEFAULT_FP_LENGTH = 2**32
DEFAULT_FOLDED_FP_LENGTH = 4096
DEFAULT_FP_DTYPE = np.int32


class Fingerprint:
    """A fingerprint that stores indices of "on" bits.

    Parameters
    ----------
    indices : array_like of int
        Indices of "on" bits.
    fp_length : int
        The fingerprint length (total number of bits).
        The default value is :math:`2^{32}`.
    unfolded_fp : `Fingerprint` or None
        The unfolded version of this fingerprint.
        If None, this fingerprint may have not been folded yet.
    unfolding_map : dict, optional
        A mapping between current indices and indices from the unfolded version
        of this fingerprint what makes it possible to trace folded bits back to
        the original shells (features).
    props: dict, optional
        Custom properties of the fingerprint, consisting of a string keyword
        and some value. It can be used, for instance, to save the ligand name
        and parameters used to generate shells (IFP features).
    """

    def __init__(self,
                 indices,
                 fp_length=DEFAULT_FP_LENGTH,
                 unfolded_fp=None,
                 unfolding_map=None,
                 props=None):

        indices = np.asarray(indices, dtype=np.int_)

        if np.any(np.logical_or(indices < 0, indices >= fp_length)):
            error_msg = "Provided indices are in a different bit scale."
            logger.exception(error_msg)
            raise BitsValueError(error_msg)

        self._indices = np.unique(indices)
        self._fp_length = fp_length
        self._unfolded_fp = unfolded_fp
        self._unfolding_map = unfolding_map or {}
        self._props = props or {}

        self.version = __version__

    @classmethod
    def from_indices(cls, indices, fp_length=DEFAULT_FP_LENGTH, **kwargs):
        """Initialize from an array of indices.

        Parameters
        ----------
        indices : array_like of int
            Indices of "on" bits.
        fp_length : int
            The fingerprint length (total number of bits).
            The default value is :math:`2^{32}`.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `Fingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import Fingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> on_bits = 8
        >>> fp_length = 32
        >>> indices = np.random.randint(0, fp_length, on_bits)
        >>> print(indices)
        [12 15 21  0  3 27  3  7]
        >>> fp = Fingerprint.from_indices(indices, fp_length=fp_length)
        >>> print(fp.indices)
        [ 0  3  7 12 15 21 27]
        >>> print(fp.to_vector(compressed=False))
        [1 0 0 1 0 0 0 1 0 0 0 0 1 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0]
        """
        return cls(indices, fp_length, **kwargs)

    @classmethod
    def from_vector(cls, vector, fp_length=None, **kwargs):
        """Initialize from a vector.

        Parameters
        ----------
        vector : :class:`numpy.ndarray` or `scipy.sparse.csr_matrix`
            Array of bits.
        fp_length : int, optional
            The fingerprint length (total number of bits).
            If not provided, the fingerprint length will be defined based on
            the ``vector`` shape.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `Fingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import Fingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> fp_length = 32
        >>> vector = np.random.choice([0, 1], size=(fp_length,), p=[0.8, 0.2])
        >>> print(vector)
        [0 0 0 0 0 0 0 1 1 0 0 0 0 1 0 0 0 1 0 1 1 0 0 0 0 0 0 1 0 0 0 0]
        >>> fp = Fingerprint.from_vector(vector)
        >>> print(fp.indices)
        [ 7  8 13 17 19 20 27]
        >>> print(fp.fp_length)
        32
        """
        if fp_length is None:
            try:
                fp_length = vector.shape[1]
            except IndexError:
                fp_length = vector.shape[0]

        if issparse(vector):
            indices = vector.indices.astype(np.int_)
        else:
            indices = np.asarray(np.where(vector), dtype=np.int_).flatten()

        return cls.from_indices(indices, fp_length, **kwargs)

    @classmethod
    def from_bit_string(cls, bit_string, fp_length=None, **kwargs):
        """Initialize from a bit string (e.g. '0010100110').

        Parameters
        ----------
        bit_string : str
            String of 0s and 1s.
        fp_length : int, optional
            The fingerprint length (total number of bits).
            If not provided, the fingerprint length will be defined based on
            the string length.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `Fingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import Fingerprint
        >>> fp = Fingerprint.from_bit_string("0010100110000010")
        >>> print(fp.indices)
        [ 2  4  7  8 14]
        >>> print(fp.fp_length)
        16
        """
        indices = [i for i, char in enumerate(bit_string) if char != '0']
        if fp_length is None:
            fp_length = len(bit_string)

        return cls.from_indices(indices, fp_length, **kwargs)

    @classmethod
    def from_rdkit(cls, rdkit_fp, **kwargs):
        """Initialize from an RDKit fingerprint.

        Parameters
        ----------
        rdkit_fp : :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or\
                        :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
            An existing RDKit fingerprint.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `Fingerprint`
        """
        if (not (isinstance(rdkit_fp, ExplicitBitVect)
                 or isinstance(rdkit_fp, SparseBitVect))):
            error_msg = ("Invalid fingerprint type. RDKit only accepts a "
                         "SparseBitVect or ExplicitBitVect object.")
            logger.exception(error_msg)
            raise TypeError(error_msg)

        fp_length = rdkit_fp.GetNumBits()
        indices = np.asarray(rdkit_fp.GetOnBits(), dtype=np.int_)

        return cls.from_indices(indices, fp_length, **kwargs)

    @classmethod
    def from_fingerprint(cls, fp, **kwargs):
        """Initialize from an existing fingerprint.

        Parameters
        ----------
        fp : `Fingerprint`
            An existing fingerprint.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `Fingerprint`
        """
        if not isinstance(fp, Fingerprint):
            error_msg = ("Informed fingerprint is not an instance of %s."
                         % (cls.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        unfolded_fp = (fp.__class__.from_fingerprint(fp.unfolded_fp)
                       if fp.unfolded_fp is not None else None)
        unfolding_map = dict(fp.unfolding_map)
        props = dict(fp.props)

        return cls.from_indices(fp.indices,
                                fp.fp_length,
                                unfolded_fp=unfolded_fp,
                                unfolding_map=unfolding_map,
                                props=props)

    @property
    def indices(self):
        """array_like of int, read-only: Indices of "on" bits."""
        return self._indices

    @property
    def bit_count(self):
        """int, read-only: Number of "on" bits."""
        return self.indices.shape[0]

    @property
    def density(self):
        """float, read-only: Proportion of "on" bits in fingerprint."""
        return self.bit_count / self.fp_length

    @property
    def counts(self):
        """dict, read-only: Mapping between each index in ``indices`` to the \
        number of counts, which is always 1 for bit fingerprints."""
        return dict([(k, 1) for k in self.indices])

    @property
    def fp_length(self):
        """int, read-only: The fingerprint length (total number of bits)."""
        return self._fp_length

    @property
    def unfolded_fp(self):
        """`Fingerprint` or None, read-only: The unfolded version of this \
        fingerprint. If None, this fingerprint may have not been folded yet."""
        if self._unfolded_fp is None:
            logger.warning("This fingerprint was not previously folded.")
            return None
        return self._unfolded_fp

    @property
    def unfolded_indices(self):
        """array_like of int, read-only: Indices of "on" bits in the unfolded \
        fingerprint."""
        if self._unfolding_map is None:
            logger.warning("This fingerprint was not previously folded.")
            return None
        return self.unfolded_fp.indices

    @property
    def unfolding_map(self):
        """dict, read-only: The mapping between current indices and indices \
        from the unfolded version of this fingerprint what makes it possible \
        to trace folded bits back to the original shells (features)."""
        if self._unfolding_map is None:
            logger.warning("This fingerprint was not previously folded.")
            return None
        return self._unfolding_map

    @property
    def props(self):
        """dict, read-only: The custom properties of the fingerprint."""
        return self._props

    @property
    def name(self):
        """str: The property 'name'. If it was not provided, then return an \
        empty string."""
        return self.props.get("name", "")

    @name.setter
    def name(self, name):
        self.props["name"] = str(name)

    @property
    def num_levels(self):
        """int: The property `num_levels` used to generate this fingerprint
        (see :class:`~luna.interaction.fp.shell.ShellGenerator`).
        If it was not provided, then return None."""
        return self.props.get("num_levels", None)

    @num_levels.setter
    def num_levels(self, num_levels):
        self.props["num_levels"] = str(num_levels)

    @property
    def radius_step(self):
        """float: The property 'radius_step' used to generate this
        fingerprint (see \
        :class:`~luna.interaction.fp.shell.ShellGenerator`). \
        If it was not provided, then return None."""
        return self.props.get("radius_step", None)

    @radius_step.setter
    def radius_step(self, radius_step):
        self.props["radius_step"] = str(radius_step)

    @property
    def num_shells(self):
        """int: The property `num_shells`
        (see :class:`~luna.interaction.fp.shell.ShellGenerator`).
        If it was not provided, then return None."""
        return self.props.get("num_shells", None)

    @num_shells.setter
    def num_shells(self, num_shells):
        self.props["num_shells"] = str(num_shells)

    def get_prop(self, key):
        """Get value of the property ``key``. If not set, raise KeyError."""
        try:
            return self.props[key]
        except KeyError:
            logger.warning("Key '%s' does not exist." % key)
            return None

    def set_prop(self, key, value):
        """Set value to the property ``key``."""
        self.props[key] = value

    def get_num_bits(self):
        """Get the fingerprint length (total number of bits)."""
        return self.fp_length

    def get_num_on_bits(self):
        """Get the number of "on" bits."""
        return self.bit_count

    def get_num_off_bits(self):
        """Get the number of "off" bits."""
        return self.get_num_bits() - self.get_num_on_bits()

    def get_bit(self, index):
        """Get the bit/count value at index ``index``.

        Raises
        ------
        BitsValueError
            If the provided index is in a different bit scale.
        """
        if index in self.counts:
            return self.counts[index]
        elif index >= 0 and index < self.fp_length:
            return 0
        else:
            error_msg = "The provided index is in a different bit scale."
            logger.exception(error_msg)
            raise BitsValueError(error_msg)

    def get_on_bits(self):
        """Get "on" bits.

        Returns
        -------
         : :class:`numpy.ndarray`
        """
        return np.array([k for (k, v) in self.counts.items() if v > 0])

    def to_vector(self, compressed=True, dtype=DEFAULT_FP_DTYPE):
        """Convert this fingerprint to a vector of bits/counts.

        .. warning::
            This function may raise a `MemoryError` exception when using
            huge indices vectors. If you found this issue, you may want
            to try a different data type or apply a folding operation
            before calling `to_vector`.

        Parameters
        -------
        compressed : bool
            If True, build a compressed sparse matrix
            (scipy.sparse.csr_matrix).
        dtype : data-type
            The default value is np.int32.

        Returns
        -------
         : :class:`numpy.ndarray` or `scipy.sparse.csr_matrix`
            Vector of bits/counts.
            Return a compressed sparse matrix (`scipy.sparse.csr_matrix`)
            if ``compressed`` is True. Otherwise, return a Numpy array
            (:class:`numpy.ndarray`)

        Raises
        ------
        BitsValueError
            If some of the fingerprint indices are greater than the
            fingerprint length.
        MemoryError
            If the operation ran out of memory.
        """
        data = [self.counts[i] for i in self.indices]
        if compressed:
            try:
                row = np.zeros(self.bit_count)
                col = self.indices
                vector = csr_matrix((data, (row, col)),
                                    shape=(1, self.fp_length),
                                    dtype=dtype)
            except ValueError as e:
                logger.exception(e)
                raise BitsValueError("Sparse matrix construction failed. "
                                     "Invalid indices or input data.")
        else:
            try:
                # This function is causing a MemoryError exception
                # when using a 2**32 vector.
                vector = np.zeros(self.fp_length, dtype=dtype)
            except MemoryError as e:
                logger.exception(e)
                raise MemoryError("Huge indices vector detected. "
                                  "An operation ran out of memory. "
                                  "Use a different data type or apply a "
                                  "folding operation.")

            try:
                vector[self.indices] = data
            except IndexError as e:
                logger.exception(e)
                raise BitsValueError("Some of the provided indices are "
                                     "greater than the fingerprint length.")

        return vector

    def to_bit_vector(self, compressed=True):
        """Convert this fingerprint to a vector of bits.

        .. warning::
            This function may raise a `MemoryError` exception when using huge
            indices vectors. If you found this issue, you may want to try a
            different data type or apply a folding operation before calling
            `to_bit_vector`.

        Parameters
        -------
        compressed : bool
            If True, build a compressed sparse matrix
            (scipy.sparse.csr_matrix).

        Returns
        -------
         : :class:`numpy.ndarray` or `scipy.sparse.csr_matrix`
            Vector of bits/counts.
            Return a compressed sparse matrix (`scipy.sparse.csr_matrix`)
            if ``compressed`` is True. Otherwise, return a Numpy array
            (:class:`numpy.ndarray`)

        Raises
        ------
        BitsValueError
            If some of the fingerprint indices are greater than the
            fingerprint length.
        MemoryError
            If the operation ran out of memory.
        """
        return self.to_vector(compressed=compressed,
                              dtype=np.bool_).astype(np.int8)

    def to_bit_string(self):
        """Convert this fingerprint to a string of bits.

        .. warning::
            This function may raise a `MemoryError` exception when using huge
            indices vectors. If you found this issue, you may want to try a
            different data type or apply a folding operation before calling
            `to_bit_string`.

        Returns
        -------
         : str

        Raises
        ------
        MemoryError
            If the operation ran out of memory.
        """
        try:
            # This function is causing a MemoryError exception
            # when using a 2**32 vector.
            bit_vector = self.to_bit_vector(compressed=False).astype(np.int8)
            return "".join(map(str, bit_vector))
        except MemoryError as e:
            logger.exception(e)
            raise MemoryError("Huge indices vector detected. An operation ran "
                              "out of memory. Use a different data type or "
                              "apply a folding operation.")

    def to_rdkit(self, rdkit_fp_cls=None):
        """Convert this fingerprint to an RDKit fingerprint.

        .. note::
            If the fingerprint length exceeds the maximum RDKit fingerprint
            length (:math:`2^{31} - 1`), this fingerprint will be folded to
            length :math:`2^{31} - 1` before conversion.

        Returns
        -------
         : :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` or \
                :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
            If ``fp_length`` is less than :math:`1e5`,
            :class:`~rdkit.DataStructs.cDataStructs.ExplicitBitVect` is used.
            Otherwise, :class:`~rdkit.DataStructs.cDataStructs.SparseBitVect`
            is used.
        """
        if rdkit_fp_cls is None:
            # Classes to store explicit bit vectors: ExplicitBitVect or
            #    SparseBitVect.
            # ExplicitBitVect is most useful for situations where the size of
            # the vector is relatively small (tens of thousands or smaller).
            # For larger vectors, use the _SparseBitVect_ class instead.
            if self.fp_length < 1e5:
                rdkit_fp_cls = ExplicitBitVect
            else:
                rdkit_fp_cls = SparseBitVect

        # RDKit data structure defines fingerprints as a std:set composed of
        # ints (signed int). Since we always have values higher than 0 and
        # since the data structure contains only signed ints, then the max
        # length for a RDKit fingerprint is 2^31 - 1.
        # C signed int (32 bit) ranges: [-2^31, 2^31-1].
        max_rdkit_fp_length = 2**31 - 1
        fp_length = self.fp_length
        if max_rdkit_fp_length < fp_length:
            logger.warning("The current fingerprint will be folded as its "
                           "size is higher than the maximum size accepted "
                           "by RDKit, which is 2**31 - 1.")
            fp_length = max_rdkit_fp_length
        indices = self.indices % max_rdkit_fp_length

        rdkit_fp = rdkit_fp_cls(fp_length)
        rdkit_fp.SetBitsFromList(indices.tolist())
        return rdkit_fp

    def fold(self, new_length=DEFAULT_FOLDED_FP_LENGTH):
        """Fold this fingerprint to size ``new_length``.

        Parameters
        ----------
        new_length : int
            Length of the new fingerprint, ideally multiple of 2.
            The default value is 4096.

        Returns
        -------
         : `Fingerprint`
            Folded `Fingerprint`.

        Raises
        ------
        BitsValueError
            If the new fingerprint length is not a multiple of 2 or is greater
            than the existing fingerprint length.

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import Fingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> on_bits = 8
        >>> fp_length = 32
        >>> indices = np.random.randint(0, fp_length, on_bits)
        >>> print(indices)
        [12 15 21  0  3 27  3  7]
        >>> fp = Fingerprint.from_indices(indices, fp_length=fp_length)
        >>> print(fp.indices)
        [ 0  3  7 12 15 21 27]
        >>> print(fp.to_vector(compressed=False))
        [1 0 0 1 0 0 0 1 0 0 0 0 1 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0]
        >>> folded_fp = fp.fold(8)
        >>> print(folded_fp.indices)
        [0 3 4 5 7]
        >>> print(folded_fp.to_vector(compressed=False))
        [1 0 0 1 1 1 0 1]
        """
        if new_length > self.fp_length:
            error_msg = ("The new fingerprint length must be smaller than the "
                         "existing fingerprint length.")
            logger.exception(error_msg)
            raise BitsValueError(error_msg)

        if not np.log2(self.fp_length / new_length).is_integer():
            error_msg = ("It is not possible to fold the current fingerprint "
                         "into the informed new length. The current length "
                         "divided by the new one is not a power of 2 number.")
            logger.exception(error_msg)
            raise BitsValueError(error_msg)

        folded_indices = self.indices % new_length

        unfolding_map = defaultdict(set)
        for k, v in sorted(zip(folded_indices, self.indices)):
            unfolding_map[k].add(v)

        props = dict(self.props)
        if "fp_length" in props:
            props["fp_length"] = new_length
        new_fp = self.__class__(indices=folded_indices,
                                fp_length=new_length,
                                unfolded_fp=self,
                                unfolding_map=unfolding_map,
                                props=props)

        return new_fp

    def unfold(self):
        """Unfold this fingerprint and return its parent fingerprint.

        Returns
        -------
         : `Fingerprint`
        """
        return self.unfolded_fp

    def union(self, other):
        """Return the union of indices of two fingerprints.

        Returns
        -------
         : :class:`numpy.ndarray`

        Raises
        ------
        InvalidFingerprintType
            If the informed fingerprint is not an instance of `Fingerprint`.
        BitsValueError
            If the fingerprints have different lengths.
        """
        if not isinstance(other, Fingerprint):
            error_msg = ("The informed fingerprint is not an instance of %s."
                         % (other.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        if self.fp_length != other.fp_length:
            raise BitsValueError("Fingerprints are in a different bit scale")

        return np.union1d(self.indices, other.indices)

    def intersection(self, other):
        """Return the intersection between indices of two fingerprints.

        Returns
        -------
         : :class:`numpy.ndarray`

        Raises
        ------
        InvalidFingerprintType
            If the informed fingerprint is not an instance of `Fingerprint`.
        BitsValueError
            If the fingerprints have different lengths.
        """
        if not isinstance(other, Fingerprint):
            error_msg = ("Informed fingerprint is not an instance of %s."
                         % (other.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        if self.fp_length != other.fp_length:
            raise BitsValueError("Fingerprints are in a different bit scale")

        return np.intersect1d(self.indices, other.indices, assume_unique=True)

    def difference(self, other):
        """Return indices in this fingerprint but not in ``other``.

        Returns
        -------
         : :class:`numpy.ndarray`

        Raises
        ------
        InvalidFingerprintType
            If the informed fingerprint is not an instance of `Fingerprint`.
        BitsValueError
            If the fingerprints have different lengths.
        """
        if not isinstance(other, Fingerprint):
            error_msg = ("Informed fingerprint is not an instance of %s."
                         % (other.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        if self.fp_length != other.fp_length:
            raise BitsValueError("Fingerprints are in a different bit scale")

        return np.setdiff1d(self.indices, other.indices, assume_unique=True)

    def symmetric_difference(self, other):
        """Return indices in either this fingerprint or ``other`` but not both.

        Returns
        -------
         : :class:`numpy.ndarray`

        Raises
        ------
        InvalidFingerprintType
            If the informed fingerprint is not an instance of `Fingerprint`.
        BitsValueError
            If the fingerprints have different lengths.
        """
        if not isinstance(other, Fingerprint):
            error_msg = ("Informed fingerprint is not an instance of %s."
                         % (other.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        if self.fp_length != other.fp_length:
            raise BitsValueError("Fingerprints are in a different bit scale")

        return np.setxor1d(self.indices, other.indices, assume_unique=True)

    def calc_similarity(self, other):
        """Calculates the Tanimoto similarity between this fingeprint
        and ``other``.

        Returns
        -------
         : float

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import Fingerprint
        >>> fp1 = Fingerprint.from_bit_string("0010101110000010")
        >>> fp2 = Fingerprint.from_bit_string("1010100110010010")
        >>> print(fp1.calc_similarity(fp2))
        0.625
        """
        return DataStructs.FingerprintSimilarity(self.to_rdkit(),
                                                 other.to_rdkit())

    def __repr__(self):
        return ("<%s: indices=%s length=%d>" %
                (self.__class__,
                 repr(self.indices).replace('\n', '').replace(' ', ''),
                 self.fp_length))

    def __eq__(self, other):
        if isinstance(other, Fingerprint):
            return (self.__class__ == other.__class__
                    and self.fp_length == other.fp_length
                    and np.all(np.in1d(self.indices,
                                       other.indices,
                                       assume_unique=True)))
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersection(other)

    def __sub__(self, other):
        return self.difference(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)


class CountFingerprint(Fingerprint):

    """A fingerprint that stores the number of occurrences of each index.

    Parameters
    ----------
    indices : array_like of int, optional
        Indices of "on" bits. It is optional if ``counts`` is provided.
    counts : dict, optional
        Mapping between each index in ``indices`` to the number of counts.
        If not provided, the default count value of 1 will be used instead.
    fp_length : int
        The fingerprint length (total number of bits).
        The default value is :math:`2^{32}`.
    unfolded_fp : `Fingerprint` or None
        The unfolded version of this fingerprint.
        If None, this fingerprint may have not been folded yet.
    unfolding_map : dict, optional
        A mapping between current indices and indices from the unfolded version
        of this fingerprint what makes it possible to trace folded bits back to
        the original shells (features).
    props: dict, optional
        Custom properties of the fingerprint, consisting of a string keyword
        and some value. It can be used, for instance, to save the ligand name
        and parameters used to generate shells (IFP features).
    """

    def __init__(self, indices=None, counts=None, fp_length=DEFAULT_FP_LENGTH,
                 unfolded_fp=None, unfolding_map=None, props=None):

        if indices is None and counts is None:
            logger.exception("Indices or counts must be provided.")
            raise IllegalArgumentError("Indices or counts must be provided.")

        if indices is not None:
            indices = np.asarray(indices, dtype=np.int_)

            if np.any(np.logical_or(indices < 0, indices >= fp_length)):
                error_msg = "Provided indices are in a different bit scale."
                logger.exception(error_msg)
                raise BitsValueError(error_msg)

            if counts is None:
                indices, counts = np.unique(indices, return_counts=True)
                counts = dict(zip(indices, counts))
            else:
                indices = np.unique(indices)
                if not np.all([x in indices for x in counts]):
                    error_msg = ("At least one index from 'counts' is not in "
                                 "'indices'.")
                    logger.exception(error_msg)
                    raise FingerprintCountsError(error_msg)

                if len(set(indices).symmetric_difference(counts)) > 0:
                    error_msg = ("At least one index in 'indices' is not "
                                 "in 'counts'.")
                    logger.exception(error_msg)
                    raise FingerprintCountsError(error_msg)
        else:
            indices = np.asarray(sorted(counts.keys()), dtype=np.int_)

            if np.any(np.logical_or(indices < 0, indices >= fp_length)):
                error_msg = ("Provided indices are in a different bit scale.")
                logger.exception(error_msg)
                raise BitsValueError(error_msg)

        self._counts = counts
        super().__init__(indices, fp_length, unfolded_fp, unfolding_map, props)

    @classmethod
    def from_indices(cls,
                     indices=None,
                     counts=None,
                     fp_length=DEFAULT_FP_LENGTH,
                     **kwargs):
        """Initialize from an array of indices.

        Parameters
        ----------
        indices : array_like of int, optional
            Indices of "on" bits. It is optional if ``counts`` is provided.
        counts : dict, optional
            Mapping between each index in ``indices`` to the number of counts.
            If not provided, the default count value of 1 will be used instead.
        fp_length : int
            The fingerprint length (total number of bits).
            The default value is :math:`2^{32}`.
        **kwargs : dict, optional
            Extra arguments to `CountFingerprint`. Refer to the documentation
            for a list of all possible arguments.

        Returns
        -------
         : `CountFingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import CountFingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> on_bits = 8
        >>> fp_length = 32
        >>> indices, counts = np.unique(np.random.randint(0, \
fp_length, on_bits), return_counts=True)
        >>> counts = dict(zip(indices, counts))
        >>> print(counts)
        {0: 1, 3: 2, 7: 1, 12: 1, 15: 1, 21: 1, 27: 1}
        >>> fp = CountFingerprint.from_indices(indices, counts=counts, \
fp_length=fp_length)
        >>> print(fp.indices)
        [ 0  3  7 12 15 21 27]
        >>> print(fp.to_vector(compressed=False))
        [1 0 0 2 0 0 0 1 0 0 0 0 1 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0]
        """
        return cls(indices=indices, 
                   counts=counts,
                   fp_length=fp_length,
                   **kwargs)

    @classmethod
    def from_counts(cls, counts, fp_length=DEFAULT_FP_LENGTH, **kwargs):
        """Initialize from a counting map.

        Parameters
        ----------
        counts : dict
            Mapping between each index in ``indices`` to the number of counts.
        fp_length : int
            The fingerprint length (total number of bits).
            The default value is :math:`2^{32}`.
        **kwargs : dict, optional
            Extra arguments to `CountFingerprint`. Refer to the documentation
            for a list of all possible arguments.

        Returns
        -------
         : `CountFingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import CountFingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> on_bits = 8
        >>> fp_length = 32
        >>> counts = dict(zip(*np.unique(np.random.randint(0, fp_length, \
on_bits),
        ...                              return_counts=True)))
        >>> print(counts)
        {0: 1, 3: 2, 7: 1, 12: 1, 15: 1, 21: 1, 27: 1}
        >>> fp = CountFingerprint.from_counts(counts=counts,
        ...                                   fp_length=fp_length)
        >>> print(fp.indices)
        [ 0  3  7 12 15 21 27]
        >>> print(fp.to_vector(compressed=False))
        1 0 0 2 0 0 0 1 0 0 0 0 1 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0]
        """
        return cls(counts=counts, fp_length=fp_length, **kwargs)

    @classmethod
    def from_bit_string(cls,
                        bit_string,
                        counts=None,
                        fp_length=None,
                        **kwargs):
        """Initialize from a bit string (e.g. '0010100110').

        Parameters
        ----------
        bit_string : str
            String of 0s and 1s.
        counts : dict, optional
            Mapping between each index in ``indices`` to the number of counts.
            If not provided, the default count value of 1 will be used instead.
        fp_length : int, optional
            The fingerprint length (total number of bits).
            If not provided, the fingerprint length will be defined based on
            the string length.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `CountFingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import CountFingerprint
        >>> fp = CountFingerprint.from_bit_string("0010100110000010",
        ...                                       counts={2: 5, 4: 1, 7: 3, \
8: 1, 14: 2})
        >>> print(fp.indices)
        [ 2  4  7  8 14]
        >>> print(fp.counts)
        {2: 5, 4: 1, 7: 3, 8: 1, 14: 2}
        """
        indices = [i for i, char in enumerate(bit_string) if char != '0']
        if fp_length is None:
            fp_length = len(bit_string)

        return cls.from_indices(indices, counts, fp_length, **kwargs)

    @classmethod
    def from_vector(cls, vector, fp_length=None, **kwargs):
        """Initialize from a vector.

        Parameters
        ----------
        vector : :class:`numpy.ndarray` or `scipy.sparse.csr_matrix`
            Array of counts.
        fp_length : int, optional
            The fingerprint length (total number of bits).
            If not provided, the fingerprint length will be defined based on
            the ``vector`` shape.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `CountFingerprint`

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import CountFingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> fp_length = 32
        >>> vector = np.random.choice(5, size=(fp_length,),
        ...                           p=[0.76, 0.1, 0.1, 0.02, 0.02])
        >>> print(vector)
        [0 0 0 0 2 3 0 1 0 0 2 0 0 0 1 1 2 3 1 0 1 0 0 0 2 0 0 0 1 0 0 0]
        >>> fp = CountFingerprint.from_vector(vector)
        >>> print(fp.indices)
        [ 4  5  7 10 14 15 16 17 18 20 24 28]
        >>> print(fp.counts)
        {4: 2, 5: 3, 7: 1, 10: 2, 14: 1, 15: 1, 16: 2, 17: 3, 18: 1, 20: 1, \
24: 2, 28: 1}
        """
        if fp_length is None:
            try:
                fp_length = vector.shape[1]
            except IndexError:
                fp_length = vector.shape[0]

        if issparse(vector):
            indices = vector.indices.astype(np.int_)
            counts = vector.data
        else:
            indices = np.asarray(np.where(vector), dtype=np.int_).flatten()
            counts = vector[indices]
        counts = dict(zip(indices, counts))

        return cls.from_indices(indices, counts, fp_length, **kwargs)

    @classmethod
    def from_fingerprint(cls, fp, **kwargs):
        """Initialize from an existing fingerprint.

        Parameters
        ----------
        fp : `Fingerprint`
            An existing fingerprint.
        **kwargs : dict, optional
            Extra arguments to `Fingerprint`. Refer to the documentation for a
            list of all possible arguments.

        Returns
        -------
         : `CountFingerprint`
        """
        if not isinstance(fp, Fingerprint):
            error_msg = ("Informed fingerprint is not an instance of %s."
                         % (cls.__class__))
            logger.exception(error_msg)
            raise InvalidFingerprintType(error_msg)

        counts = dict([(i, c) for i, c in fp.counts.items() if c > 0])
        unfolded_fp = (fp.__class__.from_fingerprint(fp.unfolded_fp)
                       if fp.unfolded_fp is not None else None)
        unfolding_map = dict(fp.unfolding_map)
        props = dict(fp.props)

        new_fp = cls.from_counts(counts, fp.fp_length, unfolded_fp=unfolded_fp,
                                 unfolding_map=unfolding_map, props=props)

        return new_fp

    @property
    def counts(self):
        """dict, read-only: Mapping between each index in ``indices`` \
        to the number of counts."""
        return self._counts

    def get_count(self, index):
        """Get the count value at index ``index``. Return 0 if index \
        is not in ``counts``."""
        return self.counts.get(index, 0)

    def fold(self, new_length=DEFAULT_FOLDED_FP_LENGTH):
        """Fold this fingerprint to size ``new_length``.

        Parameters
        ----------
        new_length : int
            Length of the new fingerprint, ideally multiple of 2.
            The default value is 4096.

        Returns
        -------
         : `Fingerprint`
            Folded `Fingerprint`.

        Raises
        ------
        BitsValueError
            If the new fingerprint length is not a multiple of 2 or is greater
            than the existing fingerprint length.

        Examples
        --------
        >>> from luna.interaction.fp.fingerprint import CountFingerprint
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> on_bits = 8
        >>> fp_length = 32
        >>> indices, counts = np.unique(np.random.randint(0, fp_length, \
on_bits),
        ...                             return_counts=True)
        >>> counts = dict(zip(indices, counts))
        >>> print(counts)
        {0: 1, 3: 2, 7: 1, 12: 1, 15: 1, 21: 1, 27: 1}
        >>> fp = CountFingerprint.from_indices(indices, counts=counts,
        ...                                    fp_length=fp_length)
        >>> print(fp.indices)
        [ 0  3  7 12 15 21 27]
        >>> print(fp.to_vector(compressed=False))
        [1 0 0 2 0 0 0 1 0 0 0 0 1 0 0 1 0 0 0 0 0 1 0 0 0 0 0 1 0 0 0 0]
        >>> folded_fp = fp.fold(8)
        >>> print(folded_fp.indices)
        [0 3 4 5 7]
        >>> print(folded_fp.to_vector(compressed=False))
        [1 0 0 3 1 1 0 2]
        """
        new_fp = super().fold(new_length)

        new_fp._counts = dict([(folded_idx, sum([self.get_count(x)
                                                 for x in unfolded_set]))
                               for folded_idx, unfolded_set
                               in new_fp.unfolding_map.items()])

        return new_fp

    def __repr__(self):
        return ("<%s: counts={%s} length=%d>" %
                (self.__class__,
                 tuple([(k, v)
                        for k, v in self.counts.items()]),
                 self.fp_length))

    def __eq__(self, other):
        if isinstance(other, Fingerprint):
            return (self.__class__ == other.__class__
                    and self.counts == other.counts
                    and self.fp_length == other.fp_length
                    and np.all(np.in1d(self.indices,
                                       other.indices,
                                       assume_unique=True)))
        return False



import numpy as np

from luna.util.math import atom_coordinates, centroid
from luna.util.default_values import PYMOL_INTERACTION_COLOR_AS_RGB
from luna.util import rgb2hex


NUCLEOPHILE_INTERS = ["Orthogonal multipolar", "Parallel multipolar",
                      "Antiparallel multipolar", "Tilted multipolar",
                      "Multipolar", "Cation-nucleophile",
                      "Unfavorable anion-nucleophile",
                      "Unfavorable nucleophile-nucleophile"]

ELECTROPHILE_INTERS = ["Orthogonal multipolar", "Parallel multipolar",
                       "Antiparallel multipolar", "Tilted multipolar",
                       "Multipolar", "Anion-electrophile",
                       "Unfavorable cation-electrophile",
                       "Unfavorable electrophile-electrophile"]

UNFAVORABLE_INTERS = ["Repulsive",
                      "Unfavorable anion-nucleophile",
                      "Unfavorable cation-electrophile",
                      "Unfavorable nucleophile-nucleophile",
                      "Unfavorable electrophile-electrophile"]


class InteractionType:
    """Define an interaction type.

    Parameters
    ----------
    src_grp : :class:`~luna.mol.groups.AtomGroup`
        The interaction’s first atom or atom group.
        For directional interactions such as hydrogen bonds, ``src_grp``
        represents donor atoms and atom groups that act as nucleophiles, i.e.,
        from where the interaction "comes from".
    trgt_grp : :class:`~luna.mol.groups.AtomGroup`
        The interaction’s second atom or atom group.
        For directional interactions such as hydrogen bonds, ``trgt_grp``
        represents acceptor atoms and atom groups that act as electrophiles,
        i.e., the interaction "receiver".
    inter_type : str
        The interaction type.
    src_interacting_atms : iterable of :class:`~luna.mol.atom.ExtendedAtom`, \
            optional
        If provided, represent the set of atoms from ``src_grp`` that in fact
        participate in the interaction. For example, in hydrophobic islands,
        not all of their atoms are in direct contact with another hydrophobic
        surface.
    trgt_interacting_atms : iterable of :class:`~luna.mol.atom.ExtendedAtom`, \
            optional
        If provided, represent the set of atoms from ``trgt_grp`` that in fact
        participate in the interaction. For example, in hydrophobic islands,
        not all of their atoms are in direct contact with another hydrophobic
        surface.
    src_centroid : array_like of float (size 3), optional
        Atomic coordinates (x, y, z) of the centroid of ``src_grp``.
        If not provided, it will be calculated automatically from ``src_grp``
        or ``src_interacting_atms``.
    trgt_centroid : array_like of float (size 3), optional
        Atomic coordinates (x, y, z) of the centroid of ``trgt_grp``.
        If not provided, it will be calculated automatically from ``trgt_grp``
        or ``trgt_interacting_atms``.
    directional : bool
        Indicate if the interaction has a direction as in hydrogen bonds and
        multipolar interactions.
    params : dict, optional
        Interaction parameters (distances, angles, etc).
    """

    def __init__(self,
                 src_grp,
                 trgt_grp,
                 inter_type,
                 src_interacting_atms=None,
                 trgt_interacting_atms=None,
                 src_centroid=None,
                 trgt_centroid=None,
                 directional=False,
                 params=None):

        self._src_grp = src_grp
        self._trgt_grp = trgt_grp

        src_interacting_atms = src_interacting_atms or []
        self._src_interacting_atms = list(src_interacting_atms)

        trgt_interacting_atms = trgt_interacting_atms or []
        self._trgt_interacting_atms = list(trgt_interacting_atms)

        self._src_centroid = (np.array(src_centroid)
                              if src_centroid is not None else None)
        self._trgt_centroid = (np.array(trgt_centroid)
                               if trgt_centroid is not None else None)

        self._type = inter_type
        self.directional = directional
        self._params = params or {}
        self._hash_cache = None

        self._apply_refs()
        self._expand_dict()

    @property
    def src_grp(self):
        """:class:`~luna.mol.groups.AtomGroup`: The interaction’s first \
                atom or atom group."""
        return self._src_grp

    @src_grp.setter
    def src_grp(self, atm_grp):
        self._src_grp = atm_grp

        # Reset hash.
        self._hash_cache = None
        self._apply_refs()

    @property
    def trgt_grp(self):
        """:class:`~luna.mol.groups.AtomGroup`: The interaction’s second atom \
                or atom group."""
        return self._trgt_grp

    @trgt_grp.setter
    def trgt_grp(self, atm_grp):
        self._trgt_grp = atm_grp

        # Reset hash.
        self._hash_cache = None
        self._apply_refs()

    @property
    def src_interacting_atms(self):
        """iterable of :class:`~luna.mol.atom.ExtendedAtom`: The set of atoms \
            from ``src_grp`` that in fact participate in the interaction. \
            If a sequence of atoms is not provided during the initialization \
            of this class, then all atoms from ``src_grp`` are returned."""
        return self._src_interacting_atms or self.src_grp.atoms

    @property
    def trgt_interacting_atms(self):
        """iterable of :class:`~luna.mol.atom.ExtendedAtom`: The set of atoms \
            from ``trgt_grp`` that in fact participate in the interaction. \
            If a sequence of atoms is not provided during the initialization \
            of this class, then all atoms from ``trgt_grp`` are returned."""
        return self._trgt_interacting_atms or self.trgt_grp.atoms

    @property
    def src_centroid(self):
        """array_like of float (size 3): Atomic coordinates (x, y, z) of the \
        centroid of ``src_grp``. If it is not provided during the \
        initialization of this class, then it will be calculated \
        automatically from ``src_grp`` or ``src_interacting_atms``.
        """
        if self._src_centroid is None:
            if self._src_interacting_atms:
                self._src_centroid = \
                    centroid(atom_coordinates(self._src_interacting_atms))
            else:
                src_centroid = self._src_grp.centroid
                # Define the centroid in a nucleophile with two atoms as the
                # position of its more electronegative atom. Remember that the
                # position in the interaction object matters. We have defined
                # that the first group is always the nucleophile for both
                # dipole-dipole and ion-dipole interactions.
                if (self.type in NUCLEOPHILE_INTERS
                        and len(self.src_grp.atoms) == 2):
                    dipole_atm = \
                        (self.src_grp.atoms[0]
                         if (self.src_grp.atoms[0].electronegativity
                             > self.src_grp.atoms[1].electronegativity)
                         else self.src_grp.atoms[1])
                    src_centroid = dipole_atm.coord

                # For unfavorable multipolar interactions, it may happen that
                # the first atom group is an electrophile as well.
                elif (self.type == "Unfavorable electrophile-electrophile"
                        and len(self.src_grp.atoms) == 2):
                    dipole_atm = \
                        (self.src_grp.atoms[0]
                         if (self.src_grp.atoms[0].electronegativity
                             < self.src_grp.atoms[1].electronegativity)
                         else self.src_grp.atoms[1])
                    src_centroid = dipole_atm.coord

                self._src_centroid = src_centroid
        return self._src_centroid

    @src_centroid.setter
    def src_centroid(self, centroid):
        if centroid is None:
            self._src_centroid = None
        else:
            self._src_centroid = np.array(centroid)

    @property
    def trgt_centroid(self):
        """array_like of float (size 3): Atomic coordinates (x, y, z) of the \
        centroid of ``trgt_grp``. If it is not provided during the \
        initialization of this class, then it will be calculated \
        automatically from ``trgt_grp`` or ``trgt_interacting_atms``.
        """
        if self._trgt_centroid is None:
            if self._trgt_interacting_atms:
                self._trgt_centroid = \
                    centroid(atom_coordinates(self._trgt_interacting_atms))
            else:
                trgt_centroid = self._trgt_grp.centroid

                # Define the centroid in an electrophile with two atoms as the
                # position of its less electronegative atom. Remember that the
                # position in the interaction object matters. We have defined
                # that the second group is always the electrophile for both
                # dipole-dipole and ion-dipole interactions.
                if (self.type in ELECTROPHILE_INTERS
                        and len(self.trgt_grp.atoms) == 2):
                    dipole_atm = \
                        (self.trgt_grp.atoms[0]
                         if (self.trgt_grp.atoms[0].electronegativity
                             < self.trgt_grp.atoms[1].electronegativity)
                         else self.trgt_grp.atoms[1])
                    trgt_centroid = dipole_atm.coord
                # For unfavorable multipolar interactions, it may happen that
                # the second atom group is a nucleophile as well.
                elif (self.type == "Unfavorable nucleophile-nucleophile"
                        and len(self.trgt_grp.atoms) == 2):
                    dipole_atm = \
                        (self.trgt_grp.atoms[0]
                         if (self.trgt_grp.atoms[0].electronegativity
                             > self.trgt_grp.atoms[1].electronegativity)
                         else self.trgt_grp.atoms[1])
                    trgt_centroid = dipole_atm.coord

                self._trgt_centroid = trgt_centroid
        return self._trgt_centroid

    @trgt_centroid.setter
    def trgt_centroid(self, centroid):
        if centroid is None:
            self._trgt_centroid = None
        else:
            self._trgt_centroid = np.array(centroid)

    @property
    def type(self):
        """str: The interaction type."""
        return self._type

    @type.setter
    def type(self, new_type):
        self._type = new_type

        # Reset hash.
        self._hash_cache = None

    @property
    def params(self):
        """dict: Interaction parameters (distances, angles, etc)."""
        return self._params

    @property
    def required_interactions(self):
        """list of `InteractionType`: If this interaction depends on other \
        interactions, then return them as a list. Currently, by default, \
        only water-bridged hydrogen bonds and salt bridges have a dependency \
        on other interactions. The first, depends on two or more \
        hydrogen bonds, while the second depends on an ionic and a \
        hydrogen bond."""

        interactions = []
        if "depends_on" in self._params:
            interactions = self._params["depends_on"]

        return interactions

    def get_partner(self, comp):
        """Get the partner atom group that forms this interaction with
        ``comp``. Return None if ``comp`` is neither the ``src_grp``
        nor ``trgt_grp``.

        Parameters
        ----------
        comp : :class:`~luna.mol.groups.AtomGroup`
            Get the partner of this atom group.

        Returns
        -------
         : :class:`~luna.mol.groups.AtomGroup` or None
        """
        if comp == self.src_grp:
            return self.trgt_grp
        elif comp == self.trgt_grp:
            return self.src_grp
        return None

    def is_directional(self):
        """Indicate if the interaction has a direction as in hydrogen bonds
        and multipolar interactions.

        Returns
        -------
         : bool
        """
        return self.directional

    def is_intramol_interaction(self):
        """Indicate if the interaction involves atoms from the same molecule.

        Returns
        -------
         : bool
        """
        comps1 = self.src_grp.compounds
        comps2 = self.trgt_grp.compounds
        return len(comps1) == 1 and len(comps2) == 1 and comps1 == comps2

    def is_intermol_interaction(self):
        """Indicate if the interaction involves atoms from different molecules.

        Returns
        -------
         : bool
        """
        return not self.is_intramol_interaction()

    def _show_src_centroid(self):
        show_centroid = True

        # Define the centroid in a nucleophile with two atoms as the position
        # of its more electronegative atom. Remember that the position in the
        # interaction object matters. We have defined that the first group is
        # always the nucleophile for both dipole-dipole and ion-dipole
        # interactions.
        if self.type in NUCLEOPHILE_INTERS and len(self.src_grp.atoms) == 2:
            show_centroid = False
        # For unfavorable multipolar interactions, it may happen that the
        # first atom group is an electrophile as well.
        elif (self.type == "Unfavorable electrophile-electrophile"
                and len(self.src_grp.atoms) == 2):
            show_centroid = False

        return show_centroid

    def _show_trgt_centroid(self):
        show_centroid = True

        # Define the centroid in an electrophile with two atoms as the position
        # of its less electronegative atom. Remember that the position in the
        # interaction object matters. We have defined that the second group is
        # always the electrophile for both dipole-dipole and ion-dipole
        # interactions.
        if self.type in ELECTROPHILE_INTERS and len(self.trgt_grp.atoms) == 2:
            show_centroid = False
        # For unfavorable multipolar interactions, it may happen that the
        # second atom group is a nucleophile as well.
        elif (self.type == "Unfavorable nucleophile-nucleophile"
                and len(self.trgt_grp.atoms) == 2):
            show_centroid = False

        return show_centroid

    def _apply_refs(self):
        self.src_grp.add_interactions([self])
        self.trgt_grp.add_interactions([self])

    def clear_refs(self):
        """References to this `InteractionType` instance will be removed from
        the list of interactions of ``src_grp`` and ``trgt_grp``."""
        self.src_grp.remove_interactions([self])
        self.trgt_grp.remove_interactions([self])

    def _expand_dict(self):
        for key in self._params:
            self.__dict__[key] = self._params[key]

    def as_json(self):
        """Represent this interaction as a dict containing the interaction
        type, flags indicating if its directional or not and if it is an
        intra- or intermolecular interaction, its default color for visual
        representations, and information related to each involved atom group.

        The dict is defined as follows:
            * ``type`` (str): the interaction type;
            * ``is_directional`` (bool`): if it is a directional interaction;
            * ``is_intramol_interaction`` (bool): if it is an intramolecular interaction;
            * ``is_intermol_interaction`` (bool): if it is an intermolecular interaction;
            * ``color`` (str): the default interaction hex color for visual representations;
            * ``src_grp`` (dict): information related to the interaction’s first atom or atom group. The dict structure is defined in :meth:`luna.mol.groups.AtomGroup.as_json`. Additional keys:

                * ``add_pseudo_group`` (bool): if True, it means a pseudo-atom (centroid) should represent ``src_grp`` as it comprises multiple atoms.
                * ``centroid`` (list of float): the atomic coordinates (x, y, z) of the centroid of ``src_grp``
                * ``show_centroid`` (str): if the centroid should be shown or not. It will be False only for nucleophiles and electrophiles.

            * ``trgt_grp`` (dict): information related to the interaction’s second atom or atom group. The dict structure is defined in :meth:`luna.mol.groups.AtomGroup.as_json`. Additional keys:

                * ``add_pseudo_group`` (bool): if True, it means a pseudo-atom (centroid) should represent ``trgt_grp`` as it comprises multiple atoms.
                * ``centroid`` (list of float): the atomic coordinates (x, y, z) of the centroid of ``trgt_grp``
                * ``show_centroid`` (str): if the centroid should be shown or not. It will be False only for nucleophiles and electrophiles.
        """
        inter_obj = {}
        inter_obj["type"] = self.type

        inter_obj["is_directional"] = self.is_directional()
        inter_obj["is_intramol_interaction"] = self.is_intramol_interaction()
        inter_obj["is_intermol_interaction"] = self.is_intermol_interaction()

        inter_obj["color"] = rgb2hex(*PYMOL_INTERACTION_COLOR_AS_RGB.get_unnormalized_color(self.type))

        src_grp_obj = self.src_grp.as_json()
        src_grp_obj["add_pseudo_group"] = len(self.src_grp.atoms) > 1
        src_grp_obj["centroid"] = self.src_centroid.tolist()
        src_grp_obj["show_centroid"] = self._show_src_centroid()

        if src_grp_obj["add_pseudo_group"]:
            src_grp_obj["pseudo_group_name"] = \
                "+".join([a.name for a in sorted(self.src_grp.atoms)])

        inter_obj["src_grp"] = src_grp_obj

        #
        # Target atom group
        #
        trgt_grp_obj = self.trgt_grp.as_json()
        trgt_grp_obj["add_pseudo_group"] = len(self.trgt_grp.atoms) > 1
        trgt_grp_obj["centroid"] = self.trgt_centroid.tolist()
        trgt_grp_obj["show_centroid"] = self._show_trgt_centroid()

        if trgt_grp_obj["add_pseudo_group"]:
            trgt_grp_obj["pseudo_group_name"] = \
                "+".join([a.name for a in sorted(self.trgt_grp.atoms)])

        inter_obj["trgt_grp"] = trgt_grp_obj

        return inter_obj

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(self, other.__class__):
            is_equal_compounds = ((self.src_grp == other.src_grp
                                   and self.trgt_grp == other.trgt_grp)
                                  or (self.src_grp == other.trgt_grp
                                      and self.trgt_grp == other.src_grp))

            is_equal_interactions = self.type == other.type
            has_equal_params = self.params == other.params

            return (is_equal_compounds and is_equal_interactions
                    and has_equal_params)
        return False

    def __ne__(self, other):
        """Overrides the default implementation"""
        return not self.__eq__(other)

    def __hash__(self):
        """Overrides the default implementation"""

        if self._hash_cache is None:
            # First, it flats the dictionary by transforming it to a list.
            # Then, it transforms the list into an immutable data structure
            # (tuple).
            params_values = []
            for key in sorted(self.params):
                if type(self.params[key]) is list:
                    val = tuple(self.params[key])
                else:
                    val = self.params[key]
                params_values.append(val)
            params_as_tuple = tuple(params_values)

            # The properties src_grp and trgt_grp properties makes an
            # InteractionType object order dependent.
            # For example, Class(X,Y) would be considered different
            # from Class(Y,X). However, in both cases the interactions
            # should be considered the same. Then, the next line turns
            # the order dependent arguments into an independent order data.
            comp_values_as_tuple = tuple(sorted([self.src_grp, self.trgt_grp]))
            self._hash_cache = hash(tuple([comp_values_as_tuple,
                                           self.type, params_as_tuple]))

        return self._hash_cache

    def __repr__(self):
        return ('<InteractionType: compounds=(%s, %s) type=%s>'
                % (self.src_grp, self.trgt_grp, self.type))



import argparse


CHARGE_METHODS = ['am1-bcc', 'gasteiger']


def main():

    parser = argparse.ArgumentParser()

    parser.add_argument('-i', dest="input", type=str, required=True,
                        help="an input molecular file.")

    parser.add_argument('-o', dest="output", type=str, required=True,
                        help="the output file.")

    parser.add_argument('-nc', dest="total_charge", type=int, required=True,
                        help="net charge.")

    parser.add_argument('-c', dest="charge_method", type=str,
                        default="am1-bcc",
                        help="Charge method is either 'am1-bcc' or "
                             "'gasteiger'. The default value is 'am1-bcc'.")

    args = parser.parse_args()

    if args.charge_method not in CHARGE_METHODS:
        raise ValueError("Unknown charge method: '%s'." % args.charge_method)

    from chimera import runCommand as rc
    from AddCharge import addNonstandardResCharges
    from chimera.selection import currentResidues

    rc("open %s" % args.input)

    rc("sel #0")  # select the ligand
    ligand = currentResidues()[0]

    addNonstandardResCharges([ligand], args.total_charge,
                             method=args.charge_method)

    rc("write format mol2 #0 %s" % args.output)

    chimera.closeSession()

    rc('stop now')


if __name__ == "__main__":
    main()



from Bio.PDB.PDBIO import Select


DEFAULT_ALTLOC = ("A", "1")


class Selector(Select):
    """
    Base selector for filtering atoms during PDB output.

    Parameters
    ----------
    keep_hydrog : bool
        If True (default), keeps all hydrogens. Otherwise, filter them out.
        Hydrogens are atoms whose element is either "H" or "D" (deuterium).
    keep_altloc : bool
        If True (default), keeps all atoms. Otherwise, keeps only atoms whose
        alternate location is in the list ``altloc``.
    altloc : iterable
        List of valid alternate location identifiers. 
        The default valid values are "A" and "1".
    """

    def __init__(self, 
                 keep_hydrog=True, 
                 keep_altloc=True, 
                 altloc=DEFAULT_ALTLOC):
        self.keep_hydrog = keep_hydrog
        self.keep_altloc = keep_altloc
        self.altloc = altloc

    def accept_atom(self, atom):
        """Return True if the atom should be written."""
        
        # Hydrogen and Deuterium
        if not self.keep_hydrog and atom.element in ["H", "D"]:
            return False
            
        if self.keep_altloc:
            return True

        return not atom.is_disordered() or atom.get_altloc() in self.altloc


class ChainSelector(Selector):
    """
    Select atoms from a specific set of chains.

    Parameters
    ----------
    entries : iterable of :class:`~luna.pdb.core.chain.Chain`
        Chains to include in the output.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_chain(self, chain):
        """Decide if the chain is valid or not."""
        return True if (chain in self.entries) else False

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return self.accept_chain(res.get_parent())

    def accept_atom(self, atom):
        """Decide if the atom is valid or not."""
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class ResidueSelector(Selector):
    """
    Select atoms from a specific set of residues.

    Parameters
    ----------
    entries : iterable of :class:`~luna.pdb.core.residue.Residue`
        Residues to include in the output.
    """
    
    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return res in self.entries

    def accept_atom(self, atom):
        """Decide if the atom is valid or not."""
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class ResidueSelectorBySeqNum(Selector):
    """
    Select residues by sequence number.

    Parameters
    ----------
    entries : iterable of int
        Residue sequence numbers to include.
    """

    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_residue(self, res):
        """Decide if the residue is valid or not."""
        return True if (res.get_id()[1] in self.entries) else False

    def accept_atom(self, atom):
        """Decide if the atom is valid or not."""
        return super().accept_atom(atom) and self.accept_residue(atom.get_parent())


class AtomSelector(Selector):
    """
    Select a specific set of atoms.

    Parameters
    ----------
    entries : iterable of :class:`~luna.pdb.core.atom.Atom`
        Atom objects to include in the output.
    """
    
    def __init__(self, entries, **kwargs):
        self.entries = entries
        super().__init__(**kwargs)

    def accept_atom(self, atom):
        return super().accept_atom(atom) and atom in self.entries



class LUNAError(Exception):
    """Base class for LUNA-specific errors.

       This class is provided for future LUNA-specific functionality.
    """
    pass


class CompatibilityError(RuntimeError):
    pass


class EntityLevelError(KeyError):
    pass


class MissingAtomsError(ValueError):
    pass


class AtomsListError(ValueError):
    pass


class ChainNotFoundError(KeyError):
    pass


class MoleculeNotFoundError(KeyError):
    pass


class IllegalArgumentError(ValueError):
    pass


class PDBNotReadError(IOError):
    pass


class PKLNotReadError(IOError):
    pass


class InvalidSuperpositionFileError(ValueError):
    pass


class FileNotCreated(IOError):
    pass


class FingerprintNotCreated(RuntimeError):
    pass


class InvalidFileFormat(ValueError):
    pass


class InvalidEntry(ValueError):
    pass


class DirectoryAlreadyExists(OSError):
    pass


class DuplicateEntry(ValueError):
    pass


class ProcessingFailed(RuntimeError):
    pass


class ShellCenterNotFound(KeyError):
    pass


class PymolSessionNotInitialized(NameError):
    pass


class InvalidFingerprintType(TypeError):
    pass


class BitsValueError(ValueError):
    pass


class FingerprintCountsError(ValueError):
    pass


class MoleculeSizeError(ValueError):
    pass


class SanitizationError(ValueError):
    pass


class AtomObjectTypeError(TypeError):
    pass


class BondObjectTypeError(TypeError):
    pass


class MoleculeObjectTypeError(TypeError):
    pass


class MoleculeObjectError(RuntimeError):
    pass


class LicenseNotFoundError(RuntimeError):
    pass



"""Consumer class that builds a Structure object.

This is used by the PDBParser and MMCIFparser classes.
"""

import warnings

from Bio.PDB.PDBExceptions import PDBConstructionException
from Bio.PDB.PDBExceptions import PDBConstructionWarning

# SMCRA hierarchy
from luna.pdb.core.structure import Structure
from luna.pdb.core.model import Model
from luna.pdb.core.chain import Chain
from luna.pdb.core.residue import Residue, DisorderedResidue
from luna.pdb.core.atom import Atom, DisorderedAtom


class StructureBuilder(object):
    """
    A Biopython-compatible structure builder for LUNA.

    Responsible for constructing the SMCRA hierarchy (Structure, Model, Chain,
    Residue, Atom) from parsed data. Includes additional features like:

    - CONECT record storage
    - Residue merging logic (for OpenBabel cleanup)
    - Disordered residue and atom handling
    - Residue indexing (idx) and line tracking
    """

    def __init__(self):
        self.line_counter = 0
        self.header = {}

        # StructureBuilder now keeps CONECT records.
        self.conects = {}

    def _is_completely_disordered(self, residue):
        """Return 1 if all atoms in the residue have a non blank altloc."""
        atom_list = residue.get_unpacked_list()
        for atom in atom_list:
            altloc = atom.get_altloc()
            if altloc == " ":
                return 0
        return 1

    # Public methods called by the Parser classes

    def set_header(self, header):
        """
        Set the PDB header block.

        Parameters
        ----------
        header : dict
            Header dictionary parsed from the PDB file.
        """
        self.header = header

    def set_conects(self, conects):
        """
        Store CONECT records parsed from the PDB file.

        Parameters
        ----------
        conects : dict
            Mapping of atom serial numbers to bonded atom serial numbers.
        """
        self.conects = conects

    def set_line_counter(self, line_counter):
        """
        Track the current line being parsed (for error reporting).

        Parameters
        ----------
        line_counter : int
            Current PDB file line number.
        """
        self.line_counter = line_counter

    def init_structure(self, structure_id, pdb_file=None):
        """
        Initialize a new Structure object.

        Parameters
        ----------
        structure_id : str
            Unique identifier for the structure.
        """
        self.structure = Structure(structure_id)

    def init_model(self, model_id, serial_num=None):
        """
        Initialize a new Model object and add it to the structure.

        Parameters
        ----------
        model_id : int
            Model index.
        serial_num : int, optional
            PDB serial number.
        """
        self.model = Model(model_id, serial_num)
        self.structure.add(self.model)

    def init_chain(self, chain_id):
        """
        Initialize a new Chain object and add it to the current model.

        Parameters
        ----------
        chain_id : str
            Chain identifier.
        """
        if self.model.has_id(chain_id):
            self.chain = self.model[chain_id]
            warnings.warn("WARNING: Chain %s is discontinuous at line %i."
                          % (chain_id, self.line_counter),
                          PDBConstructionWarning)
        else:
            self.chain = Chain(chain_id)
            self.model.add(self.chain)

    def init_seg(self, segid):
        """
        Set the current segment ID.

        Parameters
        ----------
        segid : str
            Segment ID (used by PDB files).
        """
        self.segid = segid

    def init_residue(self, resname, field, resseq, icode):
        """
        Initialize a new Residue (or DisorderedResidue) object.

        Merges with existing residue if applicable, handles alternate conformations
        and OpenBabel artifacts.

        Parameters
        ----------
        resname : str
            Residue name (e.g., "GLY").
        field : str
            Hetero flag (" ", "W", or "H" / "H_"...).
        resseq : int
            Residue sequence number.
        icode : str
            Insertion code.
        """
        if field != " ":
            if field == "H":
                # The hetero field consists of H_ + the residue name (e.g. H_FUC)
                field = "H_" + resname
        res_id = (field, resseq, icode)

        if field == " ":
            if self.chain.has_id(res_id):
                # There already is a residue with the id (field, resseq, icode).
                # This only makes sense in the case of a point mutation.
                warnings.warn("WARNING: Residue ('%s', %i, '%s') "
                              "redefined at line %i."
                              % (field, resseq, icode, self.line_counter),
                              PDBConstructionWarning)
                duplicate_residue = self.chain[res_id]
                if duplicate_residue.is_disordered() == 2:
                    # The residue in the chain is a DisorderedResidue object.
                    # So just add the last Residue object.
                    if duplicate_residue.disordered_has_id(resname):
                        # The residue was already made
                        self.residue = duplicate_residue
                        duplicate_residue.disordered_select(resname)
                    else:
                        # Make a new residue and add it to the already
                        # present DisorderedResidue
                        #
                        # Create a new residue with the new property (idx) added to the Residue class.
                        new_residue = Residue(res_id, resname, self.segid, len(self.chain.child_list), self.line_counter)
                        duplicate_residue.disordered_add(new_residue)
                        self.residue = duplicate_residue
                        return
                else:
                    if resname == duplicate_residue.resname:
                        warnings.warn("WARNING: Residue ('%s', %i, '%s','%s')"
                                      " already defined with the same name at line  %i."
                              % (field, resseq, icode, resname, self.line_counter),
                              PDBConstructionWarning)
                        self.residue = duplicate_residue
                        return
                    # Make a new DisorderedResidue object and put all
                    # the Residue objects with the id (field, resseq, icode) in it.
                    # These residues each should have non-blank altlocs for all their atoms.
                    # If not, the PDB file probably contains an error.
                    if not self._is_completely_disordered(duplicate_residue):
                        # if this exception is ignored, a residue will be missing
                        self.residue = None
                        raise PDBConstructionException(
                            "Blank altlocs in duplicate residue %s ('%s', %i, '%s')"
                            % (resname, field, resseq, icode))
                    self.chain.detach_child(res_id)

                    # Create a new residue with the new property (idx) added to the Residue class.
                    new_residue = Residue(res_id, resname, self.segid, len(self.chain.child_list), self.line_counter)
                    disordered_residue = DisorderedResidue(res_id)
                    self.chain.add(disordered_residue)
                    disordered_residue.disordered_add(duplicate_residue)
                    disordered_residue.disordered_add(new_residue)
                    self.residue = disordered_residue
                    return
        else:
            # Identify hetero residues already added to a chain.
            if self.chain.has_id(res_id):
                # There already is a hetero residue with the id (field, resseq, icode).
                # It can occur when hydrogen atoms are added in the final of the PDB file.
                warnings.warn("WARNING: Residue ('%s', %i, '%s') "
                              "redefined at line %i."
                              % (field, resseq, icode, self.line_counter),
                              PDBConstructionWarning)
                duplicate_residue = self.chain[res_id]
                if resname == duplicate_residue.resname:
                    warnings.warn("WARNING: Residue ('%s', %i, '%s','%s')"
                                  " already defined with the same name at line  %i."
                          % (field, resseq, icode, resname, self.line_counter),
                          PDBConstructionWarning)
                    self.residue = duplicate_residue
                    return

        # Create a new residue with the new property (idx) added to the Residue class.
        self.residue = Residue(res_id, resname, self.segid, len(self.chain.child_list), self.line_counter)
        self.chain.add(self.residue)

    def init_atom(self, name, coord, b_factor, occupancy, altloc, fullname,
                  serial_number=None, element=None):
        """
        Initialize a new Atom or DisorderedAtom object and add it to the current residue.

        Parameters
        ----------
        name : str
            Atom name (e.g., "CA").
        coord : np.ndarray
            Cartesian coordinates.
        b_factor : float
            B-factor.
        occupancy : float
            Occupancy value.
        altloc : str
            Alternate location indicator.
        fullname : str
            Atom name with spaces (e.g., " CA ").
        serial_number : int, optional
            Atom serial number.
        element : str, optional
            Atomic element.
        """
        residue = self.residue
        
        # if residue is None, an exception was generated during
        # the construction of the residue
        if residue is None:
            return
            
        # First check if this atom is already present in the residue.
        # If it is, it might be due to the fact that the two atoms have atom
        # names that differ only in spaces (e.g. "CA.." and ".CA.",
        # where the dots are spaces). If that is so, use all spaces
        # in the atom name of the current atom.
        if residue.has_id(name):
            duplicate_atom = residue[name]
            # atom name with spaces of duplicate atom
            duplicate_fullname = duplicate_atom.get_fullname()
            if duplicate_fullname != fullname:
                # name of current atom now includes spaces
                name = fullname
                warnings.warn("Atom names %r and %r differ "
                              "only in spaces at line %i."
                              % (duplicate_fullname, fullname,
                                 self.line_counter),
                              PDBConstructionWarning)
                
        self.atom = Atom(name, coord, b_factor, occupancy, altloc,
                         fullname, serial_number, element)
        
        if altloc != " ":
            # The atom is disordered
            if residue.has_id(name):
                # Residue already contains this atom
                duplicate_atom = residue[name]
                if duplicate_atom.is_disordered() == 2:
                    duplicate_atom.disordered_add(self.atom)
                else:
                    # This is an error in the PDB file:
                    # a disordered atom is found with a blank altloc
                    # Detach the duplicate atom, and put it in a
                    # DisorderedAtom object together with the current
                    # atom.
                    residue.detach_child(name)
                    disordered_atom = DisorderedAtom(name)
                    residue.add(disordered_atom)
                    disordered_atom.disordered_add(self.atom)
                    disordered_atom.disordered_add(duplicate_atom)
                    residue.flag_disordered()
                    warnings.warn("WARNING: disordered atom found "
                                  "with blank altloc before line %i.\n"
                                  % self.line_counter,
                                  PDBConstructionWarning)
            else:
                # The residue does not contain this disordered atom
                # so we create a new one.
                disordered_atom = DisorderedAtom(name)
                residue.add(disordered_atom)
                # Add the real atom to the disordered atom, and the
                # disordered atom to the residue
                disordered_atom.disordered_add(self.atom)
                residue.flag_disordered()
        else:
            # The atom is not disordered
            residue.add(self.atom)

    def set_anisou(self, anisou_array):
        """
        Set anisotropic B-factor for the current atom.

        Parameters
        ----------
        anisou_array : np.ndarray
            Anisotropic B-factor.
        """
        self.atom.set_anisou(anisou_array)

    def set_siguij(self, siguij_array):
        """
        Set standard deviation of anisotropic B-factor for current atom.

        Parameters
        ----------
        siguij_array : np.ndarray
        """
        self.atom.set_siguij(siguij_array)

    def set_sigatm(self, sigatm_array):
        """
        Set standard deviation of atomic positions for current atom.

        Parameters
        ----------
        sigatm_array : np.ndarray
        """
        self.atom.set_sigatm(sigatm_array)

    def get_structure(self):
        """
        Finalize and return the constructed Structure.

        Returns
        -------
        Bio.PDB.Structure
            Structure object populated from parsed data.
        """
        # Add the header dict
        self.structure.header = self.header

        # Add the conects dict.
        self.structure.conects = self.conects

        return self.structure

    def set_symmetry(self, spacegroup, cell):
        pass



# Obtained from: https://github.com/jruere/multiprocessing-logging

from __future__ import absolute_import, division, unicode_literals

import logging
import multiprocessing
import sys
import threading
import traceback


try:
    import queue
except ImportError:
    import Queue as queue  # Python 2.


# Sentinel
_SENTINEL = '__YOU_SHALL_NOT_PASS__'  


def start_mp_handler(logger=None):
    """Wraps the handlers in the given Logger with an MultiProcessingHandler.

    :param logger: whose handlers to wrap. By default, the root logger.
    """
    if logger is None:
        logger = logging.getLogger()

    for i, orig_handler in enumerate(list(logger.handlers)):
        handler = MultiProcessingHandler(
            'mp-handler-{0}'.format(i), sub_handler=orig_handler)

        logger.removeHandler(orig_handler)
        logger.addHandler(handler)


class MultiProcessingHandler(logging.Handler):

    def __init__(self, name, sub_handler=None):
        super(MultiProcessingHandler, self).__init__()

        if sub_handler is None:
            sub_handler = logging.StreamHandler()
        self.sub_handler = sub_handler

        self.setLevel(self.sub_handler.level)
        self.setFormatter(self.sub_handler.formatter)

        self.queue = multiprocessing.Queue(-1)
        self._is_closed = False
        # The thread handles receiving records asynchronously.
        self._receive_thread = threading.Thread(target=self._receive,
                                                name=name)
        self._receive_thread.daemon = True
        self._receive_thread.start()

    def setFormatter(self, fmt):
        super(MultiProcessingHandler, self).setFormatter(fmt)
        self.sub_handler.setFormatter(fmt)

    def _receive(self):
        try:
            while True:
                try:
                    record = self.queue.get()
                    if record == _SENTINEL:
                        break
                    self.sub_handler.emit(record)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except EOFError:
                    break
                except Exception:
                    traceback.print_exc(file=sys.stderr)
        finally:
            self.queue.close()
            self.queue.join_thread()

    def _send(self, s):
        self.queue.put_nowait(s)

    def _format_record(self, record):
        # ensure that exc_info and args
        # have been stringified. Removes any chance of
        # unpickleable things inside and possibly reduces
        # message size sent over the pipe.
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self._send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            self.handleError(record)

    def close(self):
        if not self._is_closed:
            self._is_closed = True
            self.queue.put_nowait(_SENTINEL)
            self._receive_thread.join(5.0)  # Waits for receive queue to empty.
            self.sub_handler.close()
            super(MultiProcessingHandler, self).close()



import logging
import logging.config
import os
import colorlog
import gzip


FILE_FORMAT = '[%(asctime)s]    %(levelname)-8s %(filename)16s:%(lineno)-10d %(processName)-20s %(message)s'
CONSOLE_FORMAT = '[%(asctime)s]    %(log_color)s%(levelname)-10s %(reset)s%(filename)16s:%(lineno)-10d %(message)s'

VERBOSITY_LEVEL = {4: logging.DEBUG,
                   3: logging.INFO,
                   2: logging.WARNING,
                   1: logging.ERROR,
                   0: logging.CRITICAL}


class CompressedRotatingFileHandler(logging.handlers.RotatingFileHandler):

    # Source: http://roadtodistributed.blogspot.com/2011/04/compressed-rotatingfilehandler-for.html
    def doRollover(self):
        self.stream.close()
        if self.backupCount > 0:
            for i in range(self.backupCount - 1, 0, -1):
                sfn = "%s.%d.gz" % (self.baseFilename, i)
                dfn = "%s.%d.gz" % (self.baseFilename, i + 1)

                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self.baseFilename + ".1.gz"

            if os.path.exists(dfn):
                os.remove(dfn)
            try:
                f_in = open(self.baseFilename, 'rb')
                f_out = gzip.open(dfn, 'wb')
                f_out.writelines(f_in)
            except Exception:
                if not os.path.exists(dfn):
                    if os.path.exists(self.baseFilename):
                        os.rename(self.baseFilename, dfn)
            finally:
                if "f_out" in dir() and f_out is not None:
                    f_out.close()
                if "f_in" in dir() and f_in is not None:
                    f_in.close()
            if os.path.exists(self.baseFilename):
                os.remove(self.baseFilename)
        self.mode = 'w'
        self.stream = self._open()


def new_logging_file(filename, logging_level=logging.INFO,
                     logger_name=None, propagate=True,
                     log_format=None, mode='a'):
    # Set the LOG file at the working path
    fh = CompressedRotatingFileHandler(filename, mode, 5 * 1024 * 1024, 20)
    file_format = log_format or FILE_FORMAT
    file_formatter = logging.Formatter(file_format, '%Y-%m-%d %H:%M:%S')
    fh.setFormatter(file_formatter)
    fh.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    console_format = log_format or CONSOLE_FORMAT
    console_formatter = \
        colorlog.ColoredFormatter(console_format, '%Y-%m-%d %H:%M:%S',
                                  log_colors={'DEBUG': 'cyan',
                                              'INFO': 'green',
                                              'WARNING': 'yellow',
                                              'ERROR': 'red',
                                              'CRITICAL': 'bold_red'})
    ch.setFormatter(console_formatter)
    ch.setLevel(logging_level)

    logger = logging.getLogger(logger_name)

    if logger_name is not None:
        logger.propagate = propagate

    # Remove the existing file handlers
    for hdlr in logger.handlers[:]:
        logger.removeHandler(hdlr)

    # Set the new handlers
    logger.addHandler(fh)
    logger.addHandler(ch)


def load_default_logging_config():
    logger = logging.getLogger()

    # Remove the existing file handlers
    for hdlr in logger.handlers[:]:
        logger.removeHandler(hdlr)

    LOGGING_CONFIG = os.path.join(os.path.dirname(__file__), "logging.cfg")
    logging.config.fileConfig(LOGGING_CONFIG)

    return logger



from pymol import cmd
from pymol import util

from luna.wrappers.cgo_arrow import cgo_arrow
from luna.util.exceptions import (PymolSessionNotInitialized,
                                  IllegalArgumentError)
from luna.util.default_values import (PYMOL_INTERACTION_COLOR,
                                      INTERACTION_SHORT_NAMES)
from luna.util.file import get_filename, get_file_format

from Bio.Data.SCOPData import protein_letters_3to1

import logging

logger = logging.getLogger()

NUCLEOPHILE_INTERS = ["Orthogonal multipolar", "Parallel multipolar",
                      "Antiparallel multipolar", "Tilted multipolar",
                      "Multipolar", "Cation-nucleophile",
                      "Unfavorable anion-nucleophile",
                      "Unfavorable nucleophile-nucleophile"]

ELECTROPHILE_INTERS = ["Orthogonal multipolar", "Parallel multipolar",
                       "Antiparallel multipolar", "Tilted multipolar",
                       "Multipolar", "Anion-electrophile",
                       "Unfavorable cation-electrophile",
                       "Unfavorable electrophile-electrophile"]

UNFAVORABLE_INTERS = ["Repulsive", "Unfavorable anion-nucleophile",
                      "Unfavorable cation-electrophile",
                      "Unfavorable nucleophile-nucleophile",
                      "Unfavorable electrophile-electrophile"]


class PymolWrapper:
    """This class provides functions to provide easy access
    to common functions from Pymol."""

    def get_cmd(self):
        """Expose the Pymol ``cmd`` object, so that one can
        call Pymol functions directly."""
        return cmd

    def load(self, input_file, obj_name=None):
        """Load a molecular file (e.g., PDB files).

        Parameters
        ----------
        input_file : str
            The pathname of the molecular file to be loaded.
        obj_name : str, optional
            Pymol object to store the loaded structure.
            If not provided, the filename will be used instead.
        """
        self.input_file = input_file
        if not obj_name:
            obj_name = get_filename(input_file)
        cmd.load(input_file, obj_name)

    def show(self, tuples):
        """Display atom and bond representations for certain selections.

        Parameters
        ----------
        tuples : iterable of tuple
            Each tuple should contain a Pymol representation (e.g., 'sticks')
            and a selection (e.g., 'hetatm').
        """
        for representation, selection in tuples:
            cmd.show(representation, selection)

    def hide(self, tuples):
        """Hide atom and bond representations for certain selections.

        Parameters
        ----------
        tuples : iterable of tuple
            Each tuple should contain a Pymol representation (e.g., 'sticks')
             and a selection (e.g., 'hetatm').
        """
        for representation, selection in tuples:
            cmd.hide(representation, selection)

    def hide_all(self):
        """Hide all representations."""
        self.hide([('everything', '')])

    def center(self, selection):
        """Translate the window, the clipping slab, and the origin to a point
        centered within the selection."""
        cmd.center(selection)

    def label(self, tuples):
        """Draw text labels for PyMOL objects.

        Parameters
        ----------
        tuples : iterable of tuple
            Each tuple should contain a selection (e.g., 'hetatm') and some
            string to label the given selection.
        """
        for selection, expression in tuples:
            cmd.label(selection, expression)

    def add_pseudoatom(self, name, opts=None):
        """Create a molecular object with a pseudoatom or add a pseudoatom to
        a molecular object if the specified object already exists.

        Parameters
        ----------
        name : str
            The object name to create or modify.
        opts : dict
            A set of options to create the pseudoatom.
            Check `Pymol <https://pymolwiki.org/index.php/Pseudoatom>`_ to
            discover which options are available.

        """
        opts = opts or {}
        cmd.pseudoatom(name, **opts)

    def select(self, selection, name="sele", enable=0):
        """Create a named selection from an atom selection.

        Parameters
        ----------
        selection : str
            The expression to select atoms.
        name : str
            The selection name, which by default is 'sele'.
        enable : {0, 1}
            If ``1``, activate the selection, i.e., show selection indicators.
            The default value is 0, which implies the selection
            indicators won't be shown.
        """
        cmd.select(name, selection, enable=enable)

    def get_names(self, obj_type):
        """Get names of objects, grouped objects, or selections.

        Parameters
        ----------
        obj_type : {'objects', 'selections', 'all', 'public_objects', \
            'public_selections', 'public_nongroup_objects', \
            'public_group_objects', 'nongroup_objects', 'group_objects'}
                The target object type.

        Returns
        -------
         : list of str
        """
        return cmd.get_names(obj_type)

    def sel_exists(self, name):
        """Check if a selection exists given by its name ``name``."""
        return name in cmd.get_names("selections")

    def obj_exists(self, name):
        """Check if an object exists given by its name ``name``."""
        return name in cmd.get_names("objects")

    def group_exists(self, name):
        """Check if a group of objects exists given by its name ``name``."""
        return name in cmd.get_names("group_objects")

    def get_coords(self, selection):
        """Get atomic coordinates for a given atom selection.

        Parameters
        ----------
        selection : str
            The expression to select atoms.

        Returns
        -------
         : array_like of float (size 3)
            Atomic coordinates (x, y, z) of each atom selected.
        """
        return cmd.get_coords(selection)

    def distance(self, name, sel1, sel2):
        """Create a new distance object between two atoms given by their
        selection-expressions.

        Parameters
        ----------
        name : str
            Name of the distance object to create.
        sel1 : str
            The expression to select the first atom.
        sel2 : str
            The expression to select the second atom.
        """
        cmd.distance(name, sel1, sel2)

    def arrow(self, name, atm_sel1, atm_sel2, opts=None):
        """Draw an arrow object between two atoms given by their
        selection-expressions.

        Parameters
        ----------
        name : str
            Name of the arrow object to create.
        sel1 : str
            The expression to select the first atom.
        sel2 : str
            The expression to select the second atom.
        opts : dict
            A set of options to create the arrow.
            Check `Pymol <https://pymolwiki.org/index.php/Cgo_arrow>`_ to
            discover which options are available.
        """
        opts = opts or {}
        cgo_arrow(atm_sel1, atm_sel2, name=name, **opts)

    def save_png(self, output_file, width=1200, height=1200, dpi=100, ray=1):
        """Save the current Pymol session as a PNG format image file.

        Parameters
        ----------
        output_file : str
            The output image pathname.
        width : int
            The width in pixels. The default value is 1,200.
        height : int or str
            The height in pixels. The default value is 1,200.
        dpi : float
            Dots-per-inch. The default value is 100.
        ray : {0, 1}
            If ``1`` (the default), run ray first to make high-resolution
            photos.
        """
        cmd.png(output_file, width, height, dpi, ray)

    def save_session(self, output_file):
        """Save the current PyMOL state to a PSE format file to later use.

        Parameters
        ----------
        output_file : str
            The output pathname.
        """
        if get_file_format(output_file) != 'pse':
            output_file += '.pse'
        cmd.save(output_file)

    def color(self, tuples):
        """Color objects and atoms.

        Parameters
        ----------
        tuples : iterable of tuple
            Each tuple should contain a color (e.g., 'red') and a selection
            (e.g., 'hetatm').
        """
        for color, selection in tuples:
            cmd.color(color, selection)

    def color_by_element(self, selections, c_color="green"):
        """Color atoms by their default element color (e.g., oxygen in red).

        Parameters
        ----------
        selections : iterable of str
            A sequence of selections to define which atoms will be colored by
            element.
        c_color : {'green', 'cyan', 'light magenta', 'yellow', 'salmon', \
                    'white', 'slate', 'bright orange', 'purple', 'pink'}
            The carbon color. The default value is 'green'.
        """

        valid_colors = ['green', 'cyan', 'light magenta', 'yellow', 'salmon',
                        'white', 'slate', 'bright orange', 'purple', 'pink']
        if c_color.lower() not in valid_colors:
            raise IllegalArgumentError("Invalid color '%s'. The accepted "
                                       "colors are: %s."
                                       % (c_color, ", ".join(valid_colors)))

        c_color = c_color.lower()

        for selection in selections:
            if c_color == 'green':
                util.cbag(selection)
            elif c_color == 'cyan':
                util.cbac(selection)
            elif c_color == 'light magenta':
                util.cbam(selection)
            elif c_color == 'yellow':
                util.cbay(selection)
            elif c_color == 'salmon':
                util.cbas(selection)
            elif c_color == 'white':
                util.cbaw(selection)
            elif c_color == 'slate':
                util.cbab(selection)
            elif c_color == 'bright orange':
                util.cbao(selection)
            elif c_color == 'purple':
                util.cbap(selection)
            elif c_color == 'pink':
                util.cbak(selection)

    def group(self, name, members, action=None):
        """Create or update a group object.

        Parameters
        ----------
        name : str
            The group name to create or update.
        members : iterable of str
            The objects to include in the group.
        action : {'add', 'remove', 'open', 'close', 'toggle', 'auto', \
                    'empty', 'purge', 'excise'}, optional
            An action to take. If not provided, the default value
            'auto' will be used instead. The description of the actions
            are described below (source: Pymol documentation):
                * add:     add members to group.
                * remove:  remove members from group (members will be \
                    ungrouped).
                * empty:   remove all members from group.
                * purge:   remove all members from group and delete them.
                * excise:  remove all members from group and delete group.
                * open:    expand group display in object menu panel.
                * close:   collapse group display in object menu panel.
                * toggle:  toggle group display in object menu panel.
                * auto:    add or toggle.
        """
        for member in members:
            if action:
                cmd.group(name, member, action)
            else:
                cmd.group(name, member)

    def create(self, name, selection):
        """Create a new molecular object from a selection.

        Note that the selected atoms won't be extracted from the original
        object. Instead, a copy of them will be created in the new object.

        Parameters
        ----------
        name : str
            The object name to be created.
        selection : str
            The expression to select atoms.
        """
        cmd.create(name, selection)

    def alter(self, selection, expression):
        """Modify atomic properties.

        Parameters
        ----------
        selection : str
            The expression to select atoms.
        expression : str
            Expression in Python language to define which properties should be
            modified. This can be used, for instance, to rename an atom or
            chain.

        Examples
        --------

        Alter the name of a ligand carbon from 'C1' to 'CA'.

        >>> pw_obj.alter("hetatm and name C1", "name='CA'")

        Alter the chain A name to 'Z'.

        >>> pw_obj.alter("chain A", "chain='Z'")
        """
        cmd.alter(selection, expression)

    def set(self, name, value, opts=None):
        """Modify global, object, object-state, or per-atom settings.

        Parameters
        ----------
        name : str
            The setting name to modify.
        value : str
            The new setting value.
        opts : dict
            A set of options.
            Check `Pymol <https://pymolwiki.org/index.php/Pseudoatom>`_
            to discover which options are available.
        """
        opts = opts or {}
        cmd.set(name, value, **opts)

    def align(self, mobile, target, opts=None):
        """Align the structure ``mobile`` to the ``target`` structure.

        Parameters
        ----------
        mobile : str
            The structure to be aligned given by an atomic selection.
        target : str
            The target structure given by an atomic selection.
        opts : dict
            Alignment options.
            Check `Pymol <https://pymolwiki.org/index.php/Align>`_
            to discover which options are available.
        """
        opts = opts or {}
        cmd.align(mobile, target, **opts)

    def delete(self, selections):
        """Delete the provided selections.

        Parameters
        ----------
        selections : iterable of str
            A sequence of selections to be deleted.
            Wildcards can be used to define object or selection names.
        """
        for selection in selections:
            cmd.delete(selection)

    def remove(self, selections):
        """Remove a selection of atoms from models.

        Parameters
        ----------
        selections : iterable of str
            A sequence of selections to define which atoms will be removed.
        """
        for selection in selections:
            cmd.remove(selection)

    def extract(self, tuples):
        """Perform multiple extractions, i.e., extract atoms from an object to
        another object.

        Parameters
        ----------
        tuples : iterable of tuple
            Each tuple should contain the object name to where atoms will be
            added and the selection itself that defines which atoms will be
            extracted (e.g., 'hetatm').
        """
        for name, selection in tuples:
            cmd.extract(name, selection)

    def load_mol_from_pdb_block(self, pdb_block, obj_name):
        """Load a molecular file from a PDB block string.

        Parameters
        ----------
        pdb_block :str
            The PDB block string.
        obj_name : str
            Pymol object to store the loaded structure.
        """
        cmd.read_pdbstr(pdb_block, obj_name, state=1)

    def reinitialize(self):
        """Clear all objects and resets all parameters to default."""
        cmd.reinitialize('everything')
        self.input_file = None

    def quit(self):
        """Terminate Pymol."""
        cmd.quit()

    def run(self, func_name, opts):
        """Run a Pymol command ``func_name`` with parameters ``opts``.

        Parameters
        ----------
        func_name : str
            The Pymol command name.
        opts : dict
            Parameters to pass to the command.
        """
        return getattr(cmd, func_name)(**opts)

    def run_cmds(self, commands):
        """Run a set of Pymol commands.

        Parameters
        ----------
        commands : iterable of tuple
            Each tuple should contain a Pymol command and its parameters.
            See :meth:`run` to more details.
        """

        for func_name, opts in commands:
            getattr(cmd, func_name)(**opts)


class PymolSessionManager:

    """Class to start, manage, and save Pymol sessions.
    This class provides useful built-in functions to load PDB/Mol files and
    show interactions.

    .. note::
        This class is not intended to be used directly because :meth:`set_view`
        is not implemented by default. Instead, you should use a class that
        inherits from `PymolSessionManager` and implements :meth:`set_view`.
        An example is the class
        :class:`~luna.interaction.view.InteractionViewer` that implements a
        custom :meth:`set_view` to show interactions. Therefore, you should
        define your own logic beyond :meth:`set_view` to save a Pymol session
        that meets your goals.

    Parameters
    ----------
    show_cartoon : bool
        If True, show the protein structure as cartoons.
    bg_color : str
        The background color. The default value is "white".
        Check `Pymol <https://pymolwiki.org/index.php/Color_Values>`_
        to discover which colors are available.
    add_directional_arrows : bool
        If True, show arrows for directional interactions (e.g., hydrogen bonds
        and multipolar interactions).
    show_res_labels : bool
        If True (the default), show residue labels.
    inter_color : :class:`~luna.util.ColorPallete`
        A Pymol-compatible color scheme for interactions.
        The default value is
        :const:`~luna.util.default_values.PYMOL_INTERACTION_COLOR`.
    pse_export_version : str
        Define a legacy format for saving Pymol sessions (PSE files).
        The default value os '1.8'.
    """

    def __init__(self,
                 show_cartoon=False,
                 bg_color="white",
                 add_directional_arrows=True,
                 show_res_labels=True,
                 inter_color=PYMOL_INTERACTION_COLOR,
                 pse_export_version="1.8"):

        self.show_cartoon = show_cartoon
        self.bg_color = bg_color
        self.pse_export_version = pse_export_version
        self.inter_color = inter_color
        self.add_directional_arrows = add_directional_arrows
        self.show_res_labels = show_res_labels
        self.wrapper = None

    def new_session(self, data, output_file):
        """Start a new session, which includes the following steps:

            * Start a new Pymol session (:meth:`start_session`);
            * Set the view (:meth:`set_view`);
            * Save the Pymol session to ``output_file``;
            * Finish the Pymol session.

        Parameters
        ----------
        data : iterable
            Data to be processed by :meth:`set_view`.
        output_file : str
            The pathname to where the Pymol session will be saved.
        """
        self.start_session()
        self.set_view(data)
        self.save_session(output_file)
        self.finish_session()

    def start_session(self):
        """Start a new session and set Pymol settings, including the
        background color and the PSE export version.
        """
        self.wrapper = PymolWrapper()
        self.wrapper.set("pse_export_version", self.pse_export_version)
        self.wrapper.set("transparency_mode", 3)
        self.wrapper.set("group_auto_mode", 2)
        self.wrapper.run_cmds([("bg_color", {"color": self.bg_color})])
        self.wrapper.set("internal_gui_width", 370)

    def set_view(self, data):
        """Set the session view. However, this method is not implemented by
        default. Instead, you should use a class that inherits from
        `PymolSessionManager` and implements :meth:`set_view`. An example is
        the class :class:`~luna.interaction.view.InteractionViewer` that
        implements a custom :meth:`set_view` to show interactions.
        Therefore, you should define your own logic beyond :meth:`set_view` to
        save a Pymol session that meets your goals.

        Parameters
        ----------
        data : iterable
            The data that will be used to set the Pymol view.
        """
        raise NotImplementedError("Use a class that implements this method.")

    def load_pdb(self,
                 pdb_file,
                 pdb_obj,
                 mol_block=None,
                 is_ftmap_output=False):
        """Load molecules from PDB files to the current Pymol session.

        Optionally, ligands can also be loaded from a separate molecular string
        block. This is especially useful when working with docked molecules in
        which the protein structure is in a PDB file and ligands are in a
        separate molecular file.

        Parameters
        ----------
        pdb_file : str
            The pathname of the PDB file to be loaded.
        pdb_obj : str
            Pymol object to store the loaded structure.
        mol_block : str, optional
            A molecular string block to load together with the PDB file.
        is_ftmap_output : bool
            If the PDB file is an FTMap output.
            If so, an additional processing step is performed to
            standardize the loaded Pymol objects.
        """
        prot_obj = "%s.prot" % pdb_obj

        self.wrapper.load(pdb_file, prot_obj)

        if is_ftmap_output:
            objs = self.wrapper.get_names("objects")
            to_merge = []
            for obj in objs:
                if obj == "protein":
                    to_merge.append(obj)

                    sel = \
                        ("protein and not resn %s"
                         % " and not resn ".join(protein_letters_3to1.keys()))
                    self.wrapper.alter(sel, "type='HETATM'")

                elif obj.endswith(".pdb"):
                    to_merge.append(obj)
                    self.wrapper.alter(obj, "type='HETATM'")

            self.wrapper.create(prot_obj, " | ".join(to_merge))
            self.wrapper.delete(objs)

        self.wrapper.extract([("%s.hets" % pdb_obj,
                               "hetatm and %s" % prot_obj)])

        if mol_block is not None:
            self.wrapper.load_mol_from_pdb_block(mol_block,
                                                 "%s.hets" % pdb_obj)

        self.wrapper.color_by_element([pdb_obj])
        self.wrapper.hide([("everything", pdb_obj)])

        if self.show_cartoon:
            self.wrapper.show([("cartoon", pdb_obj)])

    def set_interactions_view(self,
                              interactions,
                              main_grp,
                              secondary_grp=None):
        """Display molecular interactions.

        Parameters
        ----------
        interactions : iterable of \
                :class:`~luna.interaction.type.InteractionType`
            A sequence of interactions to show.
        main_grp : str
            Main Pymol object to store atom groups.
        secondary_grp : str, optional
            Secondary Pymol object to store interactions.
            If not provided, ``main_grp`` will be used instead.
        """

        residue_selections = set()

        secondary_grp = secondary_grp or main_grp

        for i, inter in enumerate(interactions):

            #
            # Centroid 1
            #
            centroid_hash1 = hash(tuple(sorted(inter.src_interacting_atms)))
            obj1_name = "%s.centroids.%s" % (main_grp, centroid_hash1)
            centroid_obj1 = inter.src_centroid
            centroid_obj1_visible = True
            # Define the centroid in a nucleophile with two atoms as the
            # position of its more electronegative atom. Remember that
            # the position in the interaction object matters. We have
            # defined that the first group is always the nucleophile for
            # both dipole-dipole and ion-dipole interactions.
            if (inter.type in NUCLEOPHILE_INTERS
                    and len(inter.src_grp.atoms) == 2):
                dipole_atm = \
                    (inter.src_grp.atoms[0]
                     if (inter.src_grp.atoms[0].electronegativity
                         > inter.src_grp.atoms[1].electronegativity)
                     else inter.src_grp.atoms[1])
                obj1_name += "_%s" % hash(dipole_atm.name)
                centroid_obj1 = dipole_atm.coord
                centroid_obj1_visible = False
            # For unfavorable multipolar interactions, it may happen that the
            # first atom group is an electrophile as well.
            elif (inter.type == "Unfavorable electrophile-electrophile"
                    and len(inter.src_grp.atoms) == 2):
                dipole_atm = \
                    (inter.src_grp.atoms[0]
                     if (inter.src_grp.atoms[0].electronegativity
                         < inter.src_grp.atoms[1].electronegativity)
                     else inter.src_grp.atoms[1])
                obj1_name += "_%s" % hash(dipole_atm.name)
                centroid_obj1 = dipole_atm.coord
                centroid_obj1_visible = False

            #
            # Centroid 2
            #
            centroid_hash2 = hash(tuple(sorted(inter.trgt_interacting_atms)))
            obj2_name = "%s.centroids.%s" % (main_grp, centroid_hash2)
            centroid_obj2 = inter.trgt_centroid
            centroid_obj2_visible = True
            # Define the centroid in an electrophile with two atoms as
            # the position of its less electronegative atom. Remember that
            # the position in the interaction object matters. We have defined
            # that the second group is always the electrophile for both
            # dipole-dipole and ion-dipole interactions.
            if (inter.type in ELECTROPHILE_INTERS
                    and len(inter.trgt_grp.atoms) == 2):
                dipole_atm = \
                    (inter.trgt_grp.atoms[0]
                     if (inter.trgt_grp.atoms[0].electronegativity
                         < inter.trgt_grp.atoms[1].electronegativity)
                     else inter.trgt_grp.atoms[1])
                obj2_name += "_%s" % hash(dipole_atm.name)
                centroid_obj2 = dipole_atm.coord
                centroid_obj2_visible = False
            # For unfavorable multipolar interactions, it may happen that the
            # second atom group is a nucleophile as well.
            elif (inter.type == "Unfavorable nucleophile-nucleophile"
                    and len(inter.trgt_grp.atoms) == 2):
                dipole_atm = \
                    (inter.trgt_grp.atoms[0]
                     if (inter.trgt_grp.atoms[0].electronegativity
                         > inter.trgt_grp.atoms[1].electronegativity)
                     else inter.trgt_grp.atoms[1])
                obj2_name += "_%s" % hash(dipole_atm.name)
                centroid_obj2 = dipole_atm.coord
                centroid_obj2_visible = False

            # Add pseudoatoms
            if not self.wrapper.obj_exists(obj1_name):
                self.wrapper.add_pseudoatom(obj1_name,
                                            {"vdw": 1,
                                             "pos": list(centroid_obj1)})
            if not self.wrapper.obj_exists(obj2_name):
                self.wrapper.add_pseudoatom(obj2_name,
                                            {"vdw": 1,
                                             "pos": list(centroid_obj2)})

            # Set the representation for each compound in the groups involved
            # in the interaction.
            compounds = inter.src_grp.compounds.union(inter.trgt_grp.compounds)
            for compound in compounds:
                if compound.is_water() or compound.is_metal():
                    comp_repr = "sphere"
                elif compound.is_hetatm():
                    if (len(compound.child_list) == 1
                            or len([atm for atm in compound.child_list
                                    if atm.element != "H"]) == 1):
                        comp_repr = "sphere"
                    else:
                        comp_repr = "sticks"
                else:
                    comp_repr = "sticks"

                comp_sel = ("%s and %s"
                            % (main_grp, bio_to_pymol_selection(compound)))
                self.wrapper.show([(comp_repr, comp_sel)])
                carb_color = "green" if compound.is_reference() else "gray"
                self.wrapper.color([(carb_color, comp_sel + " AND elem C")])

                if compound.is_residue():
                    residue_selections.add(comp_sel)

                    # Show residue label if required.
                    if self.show_res_labels:
                        self.wrapper.label([("%s AND name CA" % comp_sel,
                                             '"%s-%s" % (resn, resi)')])

            # Check if the interaction involves the same
            # compound: intramolecular interactions.
            inter_grp = "intra" if inter.is_intramol_interaction() else "inter"

            src_grp_name = "+".join(["%s-%s-%d%s"
                                     % (c.parent.id, c.resname,
                                        c.id[1], c.id[2].strip())
                                     for c in sorted(inter.src_grp.compounds)])

            trgt_grp_name = \
                "+".join(["%s-%s-%d%s" % (c.parent.id, c.resname,
                                          c.id[1], c.id[2].strip())
                          for c in sorted(inter.trgt_grp.compounds)])

            inter_grp = ("%s.all_inters.%s.%s.i%d_%s_and_%s"
                         % (secondary_grp, inter_grp,
                            INTERACTION_SHORT_NAMES[inter.type],
                            i, src_grp_name, trgt_grp_name))

            inter_name = "%s.line" % inter_grp

            self.wrapper.distance(inter_name, obj1_name, obj2_name)
            self.wrapper.hide([("label", inter_name)])

            inter_color = self.inter_color.get_color(inter.type)

            # Set styles to the interactions.
            self.wrapper.color([(inter_color,
                                 inter_name)])

            if self.add_directional_arrows:

                if inter.type in UNFAVORABLE_INTERS:
                    arrow_name1 = "%s.arrow1" % inter_grp
                    arrow_name2 = "%s.arrow2" % inter_grp

                    square_name = "%s.block" % inter_grp

                    arrow_opts = {"radius": 0.03, "gap": 0.9,
                                  "hlength": 0.5, "hradius": 0.2,
                                  "color": inter_color}
                    square_opts = {"radius": 0.3,
                                   "gap": 1.5, "hlength": 0,
                                   "hradius": 0, "color": inter_color}

                    # Two arrows in different directions
                    self.wrapper.arrow(arrow_name1,
                                       obj1_name,
                                       obj2_name,
                                       arrow_opts)

                    if not inter.is_directional():
                        self.wrapper.arrow(arrow_name2,
                                           obj2_name,
                                           obj1_name,
                                           arrow_opts)
                    self.wrapper.arrow(arrow_name2,
                                       obj2_name,
                                       obj1_name,
                                       arrow_opts)

                    # Add a square-like object
                    self.wrapper.arrow(square_name, obj1_name,
                                       obj2_name, square_opts)

                # Add arrows over the interaction lines to
                # represent directional interactions
                elif inter.is_directional():
                    arrow_name = "%s.arrow" % inter_grp
                    arrow_opts = {"radius": 0.03, "gap": 0.9,
                                  "hlength": 0.5, "hradius": 0.2,
                                  "color": inter_color}
                    self.wrapper.arrow(arrow_name,
                                       obj1_name,
                                       obj2_name,
                                       arrow_opts)

            # If a group object contains more than one atom.
            if inter.src_grp.size > 1 and centroid_obj1_visible:
                # Add the centroids to the group "centroids" and append them
                # to the main group
                self._set_centroid_style(obj1_name)
            # Otherwise, just remove the centroid as it will not add any new
            # information (the atom represented by the centroid is already the
            # atom itself).
            else:
                self.wrapper.delete([obj1_name])

            # If a group object contains more than one atom.
            if inter.trgt_grp.size > 1 and centroid_obj2_visible:
                # Add the centroids to the group "centroids" and append them
                # to the main group
                self._set_centroid_style(obj2_name)
            # Otherwise, just remove the centroid as it will not add any new
            # information (the atom represented by the centroid is already
            # been displayed).
            else:
                self.wrapper.delete([obj2_name])

        return residue_selections

    def _set_centroid_style(self, centroid):
        # Set styles to the centroid.
        self.wrapper.hide([("nonbonded", centroid)])
        self.wrapper.show([("sphere", centroid)])
        self.wrapper.set("sphere_scale", 0.2, {"selection": centroid})

    def set_last_details_to_view(self):
        """This method can be called to apply final modifications to the
        Pymol session. In its default version, the following modifications
        are applied:
            * Dash radius for interactions is set to 0.08;
            * Labels are set to bold and their size is set to 20;
            * Atomic spheres' scale is set to 0.3;
            * Hydrogen atoms are hidden;
            * The view is centered within the visible objects.
        """
        self.wrapper.set("dash_radius", 0.08)
        self.wrapper.set("label_font_id", "13")
        self.wrapper.set("label_size", "20")
        self.wrapper.set("sphere_scale", "0.3",
                         {"selection": "visible and not name PS*"})
        self.wrapper.hide([("everything", "elem H+D")])
        self.wrapper.center("visible")

    def save_session(self, output_file):
        """Save the Pymol session as a PSE file of name ``output_file``."""
        if not isinstance(self.wrapper, PymolWrapper):
            raise PymolSessionNotInitialized("No session was initialized.")

        self.wrapper.save_session(output_file)

    def finish_session(self):
        """Clear all objects and resets all parameters to default."""
        if not isinstance(self.wrapper, PymolWrapper):
            raise PymolSessionNotInitialized("No session was initialized.")

        self.wrapper.reinitialize()
        self.wrapper = None


def bio_to_pymol_selection(entity):
    """Transform an :class:`~luna.pdb.core.entity.Entity` instance into a
    Pymol selection-expression, which can then be used to select atoms in a
    Pymol session.

    Parameters
    ----------
    entity : :class:`~luna.pdb.core.entity.Entity`
        An entity to be transformed into a Pymol selection-expression.

    Returns
    -------
     : str
        The Pymol selection-expression.

    Examples
    --------

    First, let's parse a PDB file to work with.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.pdb.parser.base import PDBParser
    >>> pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    >>> structure = pdb_parser.get_structure("Protein",
    ...                                      f"{LUNA_PATH}/tutorial/\
inputs/3QQK.pdb")

    Now, let's get the Pymol selection-expression for the chain A.

    >>> from luna.wrappers.pymol import bio_to_pymol_selection
    >>> print(bio_to_pymol_selection(structure[0]['A']))
    chain A

    Finally, we can get the Pymol selection-expression for the ligand X02.

    >>> from luna.wrappers.pymol import bio_to_pymol_selection
    >>> print(bio_to_pymol_selection(structure[0]["A"][('H_X02', 497, ' ')]))
    resn X02 AND res 497 AND chain A
    """
    params = {}
    if entity.level == 'S' or entity.level == 'M':
        params[''] = 'all'
    elif entity.level == 'C':
        params['chain'] = entity.id
    elif entity.level == 'R':
        params['resn'] = entity.resname
        params['res'] = str(entity.id[1]) + entity.id[2].strip()
        params['chain'] = entity.get_parent().id
    elif entity.level == 'A':
        residue = entity.get_parent()
        params['id'] = entity.serial_number
        params['name'] = entity.name
        params['resn'] = residue.resname
        params['res'] = str(residue.id[1]) + residue.id[2].strip()
        params['chain'] = residue.get_parent().id
    else:
        return {}

    if "resn" in params:
        # Escape characters that can generate problems with Pymol
        params['resn'] = params['resn'].replace("+", "\\+")
        params['resn'] = params['resn'].replace("-", "\\-")

    return (' AND ').join(['%s %s' % (k, str(v)) for k, v in params.items()])



from subprocess import Popen, PIPE, TimeoutExpired

from openbabel.pybel import informats, outformats

from luna.util.exceptions import (FileNotCreated, InvalidFileFormat,
                                  IllegalArgumentError, ProcessingFailed)
from luna.util.default_values import OPENBABEL
from luna.util.file import is_file_valid, validate_file, get_file_format

import logging
logger = logging.getLogger()


def _prep_opts(opts, prefix=""):
    opt_list = []
    if opts is not None:
        for key in opts:
            if key == "x" and prefix == "":
                opt_list.append("-%s%s" % (key, str(opts[key])))
            else:
                if len(key) > 1:
                    opt_list.append("--%s%s" % (prefix, key))
                else:
                    opt_list.append("-%s%s" % (prefix, key))

                if opts[key] is not None:
                    opt_list.append(str(opts[key]))
    return opt_list


def mol_to_svg(infile, output, opts=None):
    """Depict a molecule as SVG using Open Babel.

    Parameters
    ----------
    infile : str
        The pathname of a molecular file.
    output : str
        Save the SVG to this file.
    opts : dict
        A set of depiction options. Check `Open Babel <http://openbabel.org/docs/dev/FileFormats/SVG_depiction.html>`_
        to discover which options are available.

    Examples
    --------

    In this example, we will depict the molecule ZINC000007786517 as SVG.
    The following options will be used:
        * C: do not draw terminal C (and attached H) explicitly;
        * d: do not display the molecule name;
        * P: image size in pixels.

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.obabel import mol_to_svg
    >>> mol_to_svg(infile=f"{LUNA_PATH}/tutorial/inputs/ZINC000007786517.mol",
    ...            output="example.svg",
    ...            opts={"C": None, "d": None, "P": 500})

    """

    logger.debug("Trying to generate a 2D diagram to the file '%s'." % infile)

    try:
        validate_file(infile)

        opt_list = _prep_opts(opts, "x")
        args = ["obabel", infile, "-O", output] + opt_list

        p = Popen(args, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()

        svg_created = stderr.decode().strip() == "1 molecule converted"
        if not is_file_valid(output) or not svg_created:
            raise FileNotCreated("SVG diagram for '%s' not created." % infile)
        else:
            logger.debug("SVG diagram for '%s' created." % infile)
    except Exception as e:
        logger.exception(e)
        raise


def convert_molecule(mol_input, input_format=None,
                     output_file=None, output_format=None,
                     opts=None, openbabel=OPENBABEL):
    """Convert a molecular file to another format using Open Babel.

    Parameters
    ----------

    mol_input : str
        The pathname of a molecular file or a SMILES string.
    input_format : str, optional
        The molecular format of ``mol_input``.
        If not provided, the format will be defined by the file ``mol_input``
        extension. If the extension could not be identified, an
        `InvalidFileFormat` exception will be raised.
    output_file : str or None
        Save the converted molecule to this file.
        If None, return the output molecule as string.
    output_format : str, optional
        The molecular format of ``output_file``.
        If not provided, the format will be defined by the ``output_file`` when
        it is not None.
        Otherwise, it will raise an `InvalidFileFormat` exception.
    opts : dict
        A set of convertion options.
        Check `Open Babel <https://openbabel.org/docs/dev/Command-line_tools/babel.html>`_
        to discover which options are available.
    openbabel : str, optional
        The Open Babel binary location.
        If not provided, the default binary ('obabel') will be used.


    Returns
    -------
     : str or None
        Return the converted molecule as a string if ``output_file`` is None.
        Otherwise, return None.

    Raises
    ------
    InvalidFileFormat
        If the provided molecular formats are not accepted by Open Babel.

    Examples
    --------

    In this example, we will convert the molecule ZINC000007786517 from
    the format MOL to MOL2 and add hydrogens to it considering a pH of 7
    (option "p").

    >>> from luna.util.default_values import LUNA_PATH
    >>> from luna.wrappers.obabel import convert_molecule
    >>> convert_molecule(mol_input=f"{LUNA_PATH}/tutorial/inputs/\
ZINC000007786517.mol",
    ...                  output_file="example.mol2",
    ...                  opts={"p": 7})
    """
    logger.info("Molecule or file '%s' will be converted to format '%s'."
                % (mol_input, input_format))

    if is_file_valid(mol_input):
        if input_format is None:
            msg = ("Input file format not defined. "
                   "It will assume the format from the file extension.")
            logger.debug(msg)
            input_format = get_file_format(mol_input)

        if input_format not in informats:
            msg = "Input format '%s' does not exist." % input_format
            raise InvalidFileFormat(msg)

        logger.debug("Input format: %s" % input_format)

    else:
        if input_format != "smi":
            msg = ("File '%s' does not exist or is not a valid file."
                   % mol_input)
            raise OSError(msg)

    if isinstance(output_file, str):
        if output_format is None:
            msg = ("Output file format not defined. It will assume the format "
                   "by the file extension.")
            logger.debug(msg)
            output_format = get_file_format(output_file, 1)

        if output_format not in outformats:
            msg = "Output format '%s' does not exist." % output_format
            raise InvalidFileFormat(msg)

        logger.debug("Output format: %s" % output_format)

    elif output_file is not None:
        msg = "The 'output_file' should be a string or None."
        raise IllegalArgumentError(msg)

    if output_format is None:
        msg = "The output format could not be identified."
        raise IllegalArgumentError(msg)

    if input_format == "smi":
        input_list = [f'-:{mol_input}']
    else:
        input_list = ['-i', input_format, mol_input]

    if output_file is None:
        output_list = ['-o', output_format]
    else:
        output_list = ['-o', output_format, '-O', output_file]

    opts = opts or {}

    timeout = 30
    if "gen3d" in opts or "minimize" in opts:
        timeout = None

    opt_list = _prep_opts(opts)
    args = [openbabel] + input_list + output_list + opt_list

    p = Popen(args, stdout=PIPE, stderr=PIPE)
    try:
        stdout, stderr = p.communicate(timeout=timeout)
    except TimeoutExpired:
        p.kill()
        raise

    error_lines = stderr.decode().strip().split("\n")
    if (error_lines[-1] == "0 molecules converted"
            or error_lines[-1] == "0 molecule converted"):
        raise ProcessingFailed("The provided molecule could not be converted.")

    if output_file is None:
        return stdout.decode().strip()

    logger.debug("File '%s' created with success." % output_file)
